
/* buffer.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 7 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Универсальный байтовый буфер ВК-6.
 *
 */

#include "cpu.h"
#include "decomp.h"
#include "buffer.h"
#include "supp.h"


unsigned char bBuffer[256];
int iBufInd0 = 0, iBufInd1 = 0;

static int index = 0;


void BufStart(void)
{ index = iBufInd0;
}

void BufWrite(unsigned char c)
{ bBuffer[index++] = c;
  if (index>255) index = 0;
}

unsigned char BufRead(void)
{ unsigned char c = bBuffer[index++];
  if (index>255) index = 0;
  return c;
}


void R9032w(void)
/* Запись данных по индексному регистру 0 */
{ bBuffer[iBufInd0] = MkByte(RX);
}

void R9033w(void)
/* Запись данных по индексному регистру 1 */
{ bBuffer[iBufInd1] = MkByte(RX);
}

void R9034w(void)
/* Запись данных по индексному регистру 1 с автоинкрементом */
{ bBuffer[iBufInd1++] = MkByte(RX);
  if (iBufInd1>255) iBufInd1=0;
}

void R9032r(void)
/* Чтение данных по индексному регистру 0 */
{ Push(bBuffer[iBufInd0]);
}

void R9033r(void)
/* Чтение данных по индексному регистру 1 */
{ Push(bBuffer[iBufInd1]);
}

void R9034r(void)
/* Чтение данных по индексному регистру 1 с автоинкрементом */
{ Push(bBuffer[iBufInd1++]);
  if (iBufInd1>255) iBufInd1=0;
}


void R9039w(void)
/* Преобразование кода команды из памяти программ в мнемонику. */
{ const unsigned char *s;
  int pc = MkByte(RX), flgBackup = flgLat;
  BufStart();
  flgLat = 1;
  s = Decomp(&pc, 0);
  flgLat = flgBackup;
  while (*s)
    BufWrite(*s++);
  BufWrite(0);
}


void R9037w(void)
/* Преобразование числа в последовательность байтов во внутреннем формате
   ЭВМ.
 */
{ unsigned char *s = ToNumMK(RX);
  int i = 0;
  BufStart();
  while (i<8)
    BufWrite(s[i++]);
}


void R9037r(void)
/* Преобразование числа из последовательности байтов во внутреннем формате
   ЭВМ.
 */
{ static unsigned char s[8];
  int i = 0;
  BufStart();
  while (i<8)
    s[i++] = BufRead();
  Push(FromNumMK(s));
}


void R9038w(void)
/* Преобразование числа в последовательность байтов в формате FLOAT. */
{ unsigned char *s = ToFloat(RX);
  int i = 0;
  BufStart();
  while (i<4)
    BufWrite(s[i++]);
}


void R9038r(void)
/* Преобразование числа из последовательсти байтов в формате FLOAT. */
{ static unsigned char s[4];
  int i = 0;
  BufStart();
  while (i<4)
    s[i++] = BufRead();
  Push(FromFloat(s));
}


void R9035w(void)
/* Вывод числа с естественной или плавающей запятой. */
{ unsigned char *s = NumToStr(RX), c;
  BufStart();
  do BufWrite(c = *s++);
  while (c);
}


void R9036w(void)
/* Вывод числа в форматированном виде. */
{ int y = Trun1Y(8), z = TruncZ(8);
  unsigned char *s = NumToStrFix(RX, y, z), c;
  BufStart();
  do BufWrite(c = *s++);
  while (c);
}


