
/* cpu.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 3 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Блок исполнения и основные команды калькулятора ВК-6.
 *
 * Функциональный узел ЭКВМ: процессор.
 * Состоит из: устройство управления, АЛУ, регистры стека, стек возврата.
 * Функциональный узел ЭКВМ: память программ.
 */

#include <stdlib.h>
#include <math.h>

#include "cpu.h"
#include "reg.h"
#include "decomp.h"
#include "display.h"
#include "clock.h"
#include "supp.h"


real RX=0, RY=0, RZ=0, RT=0, RX1=0; /* стек                         */
unsigned char bProg[10000];         /* память программ пользователя */
static int stRets[64], iRets = 0;   /* стек возврата                */
int iBreakAdr[10], nBreaks = 0;     /* точки останова               */

int flgError = 0;
/* 0 - Всё нормально.
 * 1 - ошибка ВК-6 с объяснением в строке комментария
 * 2 - ошибка командной строки (exec.c) с объяснением через OsAlert()
 */

int flgGrad = 0, iPC = 0;
int iMode = 0;
/* 0 режим автоматической работы (F АВТ)
   1 режим выполнения программы (С/П)
   2 режим программирования (F ПРГ)
   3 режим ввода байт (Р ПРГ)
 */

static real ldPi = 3.1415926535898L, ln10 = 2.302585092994L;
static real ldShift = 0, ldMantissa = 0;
static int iPoryadok = 0, iPorZnak = 1;
int iStateInput = 0;
/*  0 ввод цифры поднимет стек
    1 ввод цифры сбросит RX
    2 ввод цифры изменит целую часть
    3 ввод цифры изменит порядок
   -1 ввод цифры изменит дробную часть
 */

static function *fnRefreshText, *fnRefreshGraph;


static int Fetch(void)
{ int c = bProg[iPC++];
  if (iPC>9999) iPC = 0;
  return c;
}


static real IntPMK(real x)
{ if (x<0) return ceil(x);
  return floor(x);
}


static real FracPMK(real x)
{ if (x<0) return x-ceil(x);
  return x-floor(x);
}


void IntoAVT(int flgE)
/* Перевести ВК-6 в автоматический режим. */
{ iMode = 0;
  flgDisplay = 1; /* Вывести экран автоматического режима. */
  flgError = flgE;
}


void TurnOn(function *f1, function *f2)
/* Включение ВК-6: начальная инициализация. */
{ int i=0;
  while (i<=9999) bProg[i++] = 255;
  iRets = 0;
  fnRefreshText = f1;
  fnRefreshGraph = f2;
  flgGrad = 0;
  flgLat = 0;
  ln10 = log(10);
  ldPi = acos(-1);
  IntoAVT(0);
  FnInit();
  CommentStoreCh(13);
  RX = RY = RZ = RT = RX1 = 0;
  iStateInput = 1;
}


void TurnOff(void)
/* Выключаем ВК-6. */
{
}


static void RefreshRX(void)
/* Обновить индикацию RX во время ввода порядка. */
{ RX = ldMantissa * pow(10,iPorZnak*iPoryadok);
}


static int Overflow(void)
/* Сообщить об ошибке выполнения и вернуть 0. */
{ if (iStateInput == 3)
    RX = ldMantissa,
    iStateInput = 1;
  IntoAVT(1);  /* "Переполнение" */
  CommentLine("\x8f\xa5\xe0\xa5\xaf\xae\xab\xad\xa5\xad\xa8\xa5");
  return 0;
}


int Guard(void)
/* Проверить число в RX на переполнение, которое может случиться
   при вводе числа из-за особенностей команды ВП.
   Возвратить 0, если произошла ошибка.
 */
{ if (Abs(RX)>MAX_REAL14) return Overflow();
  iStateInput = 0;
  flgError = 0;
  if (Abs(RX)<MIN_REAL) RX = 0;
  return 1;
}


int Guard12(void)
/* Тоже, что Guard(), но для 12-разрядной мантиссы. */
{ if (Abs(RX)>MAX_REAL12) return Overflow();
  iStateInput = 0;
  flgError = 0;
  if (Abs(RX)<MIN_REAL) RX = 0;
  return 1;
}


static void MathOverflow(void)
{ IntoAVT(1); /* "Переполнение" */
  CommentLine("\x8f\xa5\xe0\xa5\xaf\xae\xab\xad\xa5\xad\xa8\xa5");
  iStateInput = 0;
}


static void RetsOverflow(void)
{ IntoAVT(1); /* "Переполнен стек ПП" */
  CommentLine("\x8f\xa5\xe0\xa5\xaf\xae\xab\xad\xa5\xad "
    "\xe1\xe2\xa5\xaa \x8f\x8f");
  iStateInput = 0;
}


static void ErrArg(void)
{ IntoAVT(1);
  CommentLine("\x8d\xa5\xa2\xa5\xe0\xad\xeb\xa9"
    " \xa0\xe0\xa3\xe3\xac\xa5\xad\xe2"); /* "Неверный аргумент" */
  iStateInput = 0;
}


static void ErrAddr(void)
{ IntoAVT(1);
  CommentLine("\x8d\xe0\xe1\xe3\xe9\xa5\xe1\xe2\xa2\xe3\xee\xe9\xa8\xa9"
    " \xa0\xa4\xe0\xa5\xe1"); /* "Несуществующий адрес" */
  iStateInput = 0;
}


static void ErrDiv0(void)
{ IntoAVT(1);
  CommentLine("\x84\xa5\xab\xa5\xad\xa8\xa5" /* "Деление на ноль" */
    " \xad\xa0 \xad\xae\xab\xec");
  iStateInput = 0;
}


static int DestPage(int iPCref)
{ if (iPCref>9999) iPCref -= 10000;
  return (iPCref/100)*100;
}


static real GetAngle(int flgTG)
{ real q = RX, dq;
  switch (flgGrad)
  { case 0:  dq = 180;   break;
    case 1:  dq = ldPi;  break;
    case 2:  dq = 200;   break;
  }
  if (flgTG==0) dq *= 2;
  if (q<0)   return q + dq*floor(-q/dq);
  if (q>=dq) return q - dq*floor(q/dq);
  return q;
}


static real TrigArg(int flgTG)
{ real q = GetAngle(flgTG);
  switch (flgGrad)
  { case 0:  return q*ldPi/180;
    case 1:  return q;
    case 2:  return q*ldPi/200;
  }
}


static real MkAngle(real q)
{ switch (flgGrad)
  { case 0:  return q*180/ldPi;
    case 1:  return q;
    case 2:  return q*200/ldPi;
  }
}


void GRAD(void)
  { flgGrad = (flgGrad+1)%3; }

static void Oper(real x)
  { RX1 = RX;  RX = x;  RY = RZ;  RZ = RT; }


void Cx(void)
  { State1(0x0d);  RX = 0;  iStateInput = 1;  flgError = 0; }

void Enter(void)
  { if (Guard()) State1(0x0e),  Push(RX),  iStateInput = 1; }

void XCHG(void)
  { if (Guard()) State1(0x14),  RX1 = RX,  RX = RY,  RY = RX1; }

void FANS(void)
  { if (Guard()) State1(0x0f),  RT = RZ,  RZ = RY,  RY = RX,  RX = RX1; }

void FR(void)
  { if (Guard()) State1(0x25),  RX1=RX,  RX=RY,  RY=RZ,  RZ=RT,  RT=RX1; }

void FPI(void)
  { if (Guard()) State1(0x20),  RX1 = RX,  Push(ldPi); }

void Fsin(void)
  { if (Guard()) State1(0x1c),  RX1 = RX,  RX = sin(TrigArg(0)); }

void Fcos(void)
  { if (Guard()) State1(0x1d),  RX1 = RX,  RX = cos(TrigArg(0)); }

void Farctg(void)
  { if (Guard()) State1(0x1b),  RX1 = RX,  RX = MkAngle(atan(RX)); }


void KABS(void)
  { if (Guard()) State1(0x31),  RX1 = RX,  RX = Abs(RX); }

void KSGN(void)
  { if (Guard()) { State1(0x32);  RX1 = RX;  RX = (RX>0)? 1: (RX<0)? -1: 0; } }

void KNOT(void)
  { if (Guard()) State1(0x3a),  RX1 = RX,  RX = 255-MkByte(RX); }

void KRAN(void)
  { if (Guard()) State1(0x3b),  RX1 = RX,  Push(rand()/(RAND_MAX+1.L)); }

void KPRGM(void)
  { if (Guard()) State1(0x28),  RX1 = RX,  RX = bProg[Trunc(9999)]; }

void KAND(void)
  { if (Guard()) State1(0x37),  Oper(MkByte(RX) & MkByte(RY)); }

void KOR(void)
  { if (Guard()) State1(0x38),  Oper(MkByte(RX) | MkByte(RY)); }

void KXOR(void)
  { if (Guard()) State1(0x39),  Oper(MkByte(RX) ^ MkByte(RY)); }

void KINT(void)
  { if (Guard()) State1(0x34),  RX1 = RX,  RX = IntPMK(RX); }

void KFRAC(void)
  { if (Guard()) State1(0x35),  RX1 = RX,  RX = FracPMK(RX); }

void KSCR(void)
  { if (Guard()) State1(0x55),  fnRefreshText(), flgDisplay = 0; }

void KGRPH(void)
  { if (Guard()) State1(0x56),  fnRefreshGraph(), flgDisplay = 0; }

void KNOP(void)
  { if (Guard()) State1(0x54); }


static void Stop(void)
  { if (Guard()) IntoAVT(0); }

void B0(void)
  { State1(0x52);  iPC = 0;  flgError = 0;  iRets = 0; }

static void Nope(void)
  { Guard(); }


void FEXP(void)
{ if (Guard())
  { RX1 = RX;
    if (RX<=230.2585092994L)
    { State1(0x16);
      RX = exp(RX);
      if (RX<MIN_REAL) RX = 0;
    } else MathOverflow();
  }
}


void F10X(void)
{ if (Guard())
  { RX1 = RX;
    if (RX<100)
    { State1(0x15);
      RX = exp(RX*ln10);
      if (RX<MIN_REAL) RX = 0;
    } else MathOverflow();
  }
}


void FXY(void)
{ if (Guard())
  { RX1 = RX;
    if (RX<0) ErrArg();
    else if (RX!=0)
    { real z = RY*log(RX);
      if (z<=230.2585092994L)
      { State1(0x24);
        RX = exp(z);
        if (RX<MIN_REAL) RX = 0;
      } else MathOverflow();
    } else if (RY==0) ErrArg(); else State1(0x24);
  }
}


void Fln(void)
{ if (Guard())
    if (RX<=0) ErrArg();
    else State1(0x18),  RX1 = RX,  RX = log(RX);
}


void Flg(void)
{ if (Guard())
    if (RX<=0) ErrArg();
    else State1(0x17),  RX1 = RX,  RX = log10(RX);
}


void FSQRT(void)
{ if (Guard())
  { RX1 = RX;
    if (RX<0) ErrArg();
    else State1(0x21),  RX = sqrt(RX);
  }
}


void FX2(void)
{ if (Guard())
  { RX1 = RX;
    RX *= RX;
    if (RX<=MAX_REAL14)
    { State1(0x22);
      if (RX<MIN_REAL) RX = 0;
    } else RX = RX1, MathOverflow();
  }
}


void F1X(void)
{ if (Guard())
  { RX1 = RX;
    if (RX==0) ErrDiv0();
    else State1(0x23),  RX = 1.0L/RX;
  }
}


void Ftg(void)
{ if (Guard())
  { int q = GetAngle(1);
    RX1 = RX;
    if ( ((flgGrad==1) && (Abs(RX-ldPi/2)<1E-13L)) ||
         ((flgGrad==0) && (Abs(RX-90)<1E-12L)) ||
         ((flgGrad==2) && (Abs(RX-100)<1E-11L)) ) ErrArg();
    else State1(0x1e),  RX = tan(TrigArg(1));
  }
}


void Farcsin(void)
{ if (Guard())
  { RX1 = RX;
    if ((RX<-1) || (RX>1)) ErrArg();
    else State1(0x19),  RX = MkAngle(asin(RX));
  }
}


void Farccos(void)
{ if (Guard())
  { RX1 = RX;
    if ((RX<-1) || (RX>1)) ErrArg();
    else State1(0x1a), RX = MkAngle(acos(RX));
  }
}


void KMINUS(void)
{ if (Guard())
    IntoAVT(1),                   /* "Ошибка К-" */
    CommentLine("\x8e\xe8\xa8\xa1\xaa\xa0 \x8a-");
}


void KDIV(void)
{ if (Guard())
    IntoAVT(1),                      /* "Ошибка К%" */
    CommentLine("\x8e\xe8\xa8\xa1\xaa\xa0 \x8a\xb0");
}


void Kmax(void)
{ if (Guard())
  { RX1 = RX;
    if (RX<RY) RX = RY,  RY = RX1;
    State1(0x36);
  }
}


void KMD(void)
{ if (Guard())
  { real d = IntPMK(RX), m = FracPMK(RX);
    RX1 = RX;
    if (Abs(m)>=0.6L) ErrArg();
    else State1(0x26),  RX = d + m/0.6L;
  }
}


void KDM(void)
{ if (Guard())
  { real d = IntPMK(RX), f = FracPMK(RX);
    State1(0x33);
    RX1 = RX;
    RX = d + f*0.6L;
  }
}


void KMSD(void)
{ if (Guard())
  { real d = IntPMK(RX), m = FracPMK(RX), s;
    m *= 100;
    s = FracPMK(m);
    m = IntPMK(m);
    if ((Abs(s)>=0.6L) || (Abs(m)>=60)) ErrArg();
    else State1(0x2a),  RX = d + (s/0.6L + m)/60;
  }
}


void KDMS(void)
{ if (Guard())
  { real d = IntPMK(RX), f = FracPMK(RX);
    f *= 60;
    RX = d + IntPMK(f)/100 + FracPMK(f)*0.006L;
    State1(0x30);
  }
}


void Plus(void)
{ if (Guard())
  { RX1 = RX;
    RX += RY;
    if (Abs(RX)<=MAX_REAL14)
    { State1(0x10);
      if (Abs(RX)<MIN_REAL) RX = 0;
      RY = RZ;  RZ = RT;
    } else RX = RX1,  MathOverflow();
  }
}


void Minus(void)
{ if (Guard())
  { RX1 = RX;
    RX = RY-RX;
    if (Abs(RX)<=MAX_REAL14)
    { State1(0x11);
      if (Abs(RX)<MIN_REAL) RX = 0;
      RY = RZ;  RZ = RT;
    } else RX = RX1,  MathOverflow();
  }
}


void Multiply(void)
{ if (Guard())
  { RX1 = RX;
    RX *= RY;
    if (Abs(RX)<=MAX_REAL14)
    { State1(0x12);
      if (Abs(RX)<MIN_REAL) RX = 0;
      RY = RZ;  RZ = RT;
    } else RX = RX1,  MathOverflow();
  }
}


void Divide(void)
{ if (Guard())
  { RX1 = RX;
    if (RX!=0)
    { RX = RY/RX;
      if (Abs(RX)<=MAX_REAL14)
      { State1(0x13);
        if (Abs(RX)<MIN_REAL) RX = 0;
        RY = RZ;  RZ = RT;
      } else RX = RX1,  MathOverflow();
    } else ErrDiv0();
  }
}


void Digit(int x)
{ static int iEntered = 0;
  State1(x);
  switch (iStateInput)
  { case -1: if (iEntered<8)
             { iEntered++;
               ldShift /= 10.;
               RX += x*ldShift;
             }
             break;
    case 0:  Push(RX);
    case 1:  RX = 0;
             iEntered = 0;
             iStateInput = 2;
    case 2:  if (iEntered<8) iEntered++, RX = RX*10+x;
             break;
    case 3:  iPoryadok = (iPoryadok%10)*10+x;
             RefreshRX();
  }
  flgError = 0;
}


void Tochka(void)
{ switch (iStateInput)
  { case 3:  IntoAVT(1);          /* "Ошибка ВП ," */
             CommentLine("\x8e\xe8\xa8\xa1\xaa\xa0 \x82\x8f ,");
             iStateInput = 1;
             break;
    case 0:  Enter();
    case 1:  RX = 0;
    case 2:  iStateInput = -1;
             ldShift = 1.;
    default: flgError = 0;
             State1(0x0a);
  }
}


void Negate(void)
{ State1(0x0b);
  switch (iStateInput)
  { case 3:  iPorZnak = -iPorZnak;
             RefreshRX();
             break;
    case 2:
    case -1: iStateInput = 1;
    default: RX = -RX;
             break;
  }
  flgError = 0;
}


void EE(void)
{ if (Guard())
  { State1(0x0c);
    if (RX==0) RX = 1;
    ldMantissa = RX;
    iPoryadok = 0;
    iPorZnak = 1;
    iStateInput = 3;
  }
}


void GOTO(int bcd, int iPCref)
{ if (Guard())
  { int dest = DestPage(iPCref) + Bin(bcd);
    if (dest>9999) ErrAddr();
    else State(MnemoGOTO(bcd)),  iPC=dest;
  }
}


void GSB(int bcd, int iPCref)
{ if (Guard())
  { int dest = DestPage(iPCref) + Bin(bcd);
    if (dest>9999) ErrAddr();
    else if (iRets>=64) RetsOverflow();
    else stRets[iRets++] = iPC,  iPC = dest;
  }
}


void PGOTO(int bcd1, int bcd2)
{ if (Guard())
  { int dest = Bin2(bcd1, bcd2);
    if (dest>9999) ErrAddr();
    else State(MnemoPGOTO(bcd1, bcd2)),  iPC = dest;
  }
}


void PGSB(int bcd1, int bcd2)
{ if (Guard())
  { int dest = Bin2(bcd1, bcd2);
    if (dest>9999) ErrAddr();
    else if (iRets>=64) RetsOverflow();
    else stRets[iRets++] = iPC,  iPC = dest;
  }
}


void KGOTO(int r)
{ if (Guard())
  { Modify(r,1);
    if (ldR[r]>9999) ErrAddr();
    else State(MnemoKGOTO(r)),  iPC = (int) ldR[r];
  }
}


void KGSB(int r)
{ if (Guard())
  { Modify(r,1);
    if (ldR[r]>9999) ErrAddr();
    else if (iRets>=64) RetsOverflow();
    else stRets[iRets++] = iPC,  iPC = (int) ldR[r];
  }
}


void PKGOTO(int bcd)
{ if (Guard())
  { int r = Bin(bcd);
    Modify(r,0);
    if (ldR[r]>9999) ErrAddr();
    else State(MnemoPKGOTO(bcd)),  iPC = (int) ldR[r];
  }
}


void PKGSB(int bcd)
{ if (Guard())
  { int r = Bin(bcd);
    Modify(r,0);
    if (ldR[r]>9999) ErrAddr();
    else if (iRets>=64) RetsOverflow();
    else stRets[iRets++] = iPC,  iPC = (int) ldR[r];
  }
}


void FL(int r, int bcd, int iPCref)
{ if (Guard())
    if ((ldR[r]>=1) && (ldR[r]<2)) ldR[r]=1; else
    { Modify(r,1);
      if (ldR[r]>0)
      { int dest = DestPage(iPCref) + Bin(bcd);
        if (dest<=9999) iPC = dest; else ErrAddr();
      }
    }
}


void PFL(int r, int bcd1, int bcd2)
{ if (Guard())
    if ((ldR[r]>=1) && (ldR[r]<2)) ldR[r]=1; else
    { Modify(r,1);
      if (ldR[r]>0)
      { int dest = Bin2(bcd1, bcd2);
        if (dest<=9999) iPC = dest; else ErrAddr();
      }
    }
}


static void FXNE0(int bcd, int iPCref)
{ if (Guard())
    if (RX==0)
    { int dest = DestPage(iPCref) + Bin(bcd);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void KXNE0(int r)
{ if (Guard())
    if (RX==0)
    { Modify(r, 1);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void PKXNE0(int bcd)
{ if (Guard())
    if (RX==0)
    { int r = Bin(bcd);
      Modify(r, 0);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void FXEQ0(int bcd, int iPCref)
{ if (Guard())
    if (RX!=0)
    { int dest = DestPage(iPCref) + Bin(bcd);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void KXEQ0(int r)
{ if (Guard())
    if (RX!=0)
    { Modify(r, 1);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void PKXEQ0(int bcd)
{ if (Guard())
    if (RX!=0)
    { int r = Bin(bcd);
      Modify(r, 0);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void FXLT0(int bcd, int iPCref)
{ if (Guard())
    if (RX>=0)
    { int dest = DestPage(iPCref) + Bin(bcd);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void KXLT0(int r)
{ if (Guard())
    if (RX>=0)
    { Modify(r, 1);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void PKXLT0(int bcd)
{ if (Guard())
    if (RX>=0)
    { int r = Bin(bcd);
      Modify(r, 0);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void FXGE0(int bcd, int iPCref)
{ if (Guard())
    if (RX<0)
    { int dest = DestPage(iPCref) + Bin(bcd);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void KXGE0(int r)
{ if (Guard())
    if (RX<0)
    { Modify(r, 1);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void PKXGE0(int bcd)
{ if (Guard())
    if (RX<0)
    { int r = Bin(bcd);
      Modify(r, 0);
      if (ldR[r]<=9999) iPC = (int) ldR[r]; else ErrAddr();
    }
}


static void PXNE0(int bcd1, int bcd2)
{ if (Guard())
    if (RX==0)
    { int dest = Bin2(bcd1, bcd2);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void PXEQ0(int bcd1, int bcd2)
{ if (Guard())
    if (RX!=0)
    { int dest = Bin2(bcd1, bcd2);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void PXLT0(int bcd1, int bcd2)
{ if (Guard())
    if (RX>=0)
    { int dest = Bin2(bcd1, bcd2);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


static void PXGE0(int bcd1, int bcd2)
{ if (Guard())
    if (RX<0)
    { int dest = Bin2(bcd1, bcd2);
      if (dest<=9999) iPC = dest; else ErrAddr();
    }
}


void RTN(int iPCref)
{ if (Guard())
  { if (iRets>0) iPC = stRets[--iRets];
    else iPC = DestPage(iPCref) + 1;
  }
}


function *(Op[])=
{ Tochka, Negate, EE, Cx, Enter, FANS,
  Plus, Minus, Multiply, Divide, XCHG, F10X, FEXP, Flg,
  Fln, Farcsin, Farccos, Farctg, Fsin, Fcos, Ftg, Nope,
  FPI, FSQRT, FX2, F1X, FXY, FR, KMD, KMINUS,
  KPRGM, KDIV, KMSD, Nope, Nope, Nope, Nope, Nope,
  KDMS, KABS, KSGN, KDM, KINT, KFRAC, Kmax, KAND,
  KOR, KXOR, KNOT, KRAN, Nope, Nope, Nope, Nope
};


void Step(int flgSilent)
/* Выполнить один шаг программы.
   flgSilent определяет, выводить ли следующую команду в строку комментариев.
 */
{ int iPCref = iPC;
  int c = Fetch();
  TimerAdjust();
  if (c<10) Digit(c);
  else if (c<0x40) Op[c-10]();
  else if (c<=0x4e) M(c-0x40);
  else if ((c>=0x60) && (c<=0x6e)) RM(c-0x60);
  else if ((c>=0x70) && (c<=0x7e)) KXNE0(c-0x70);
  else if ((c>=0x80) && (c<=0x8e)) KGOTO(c-0x80);
  else if ((c>=0x90) && (c<=0x9e)) KXGE0(c-0x90);
  else if ((c>=0xa0) && (c<=0xae)) KGSB(c-0xa0);
  else if ((c>=0xb0) && (c<=0xbe)) KM(c-0xb0);
  else if ((c>=0xc0) && (c<=0xce)) KXLT0(c-0xc0);
  else if ((c>=0xd0) && (c<=0xde)) KRM(c-0xd0);
  else if ((c>=0xe0) && (c<=0xee)) KXEQ0(c-0xe0);
  else switch(c)
  { case 0x4f: PM(Fetch());     break;
    case 0x50: Stop();          break;
    case 0x51: iPCref = iPC;  GOTO(Fetch(), iPCref);   break;
    case 0x52: RTN(iPCref);     break;
    case 0x53: iPCref = iPC;  GSB(Fetch(), iPCref);    break;
    case 0x54: KNOP();          break;
    case 0x55: KSCR();          break;
    case 0x56: KGRPH();         break;
    case 0x57: iPCref = iPC;  FXNE0(Fetch(), iPCref);  break;
    case 0x58: iPCref = iPC;  FL(2, Fetch(), iPCref);  break;
    case 0x59: iPCref = iPC;  FXGE0(Fetch(), iPCref);  break;
    case 0x5a: iPCref = iPC;  FL(3, Fetch(), iPCref);  break;
    case 0x5b: iPCref = iPC;  FL(1, Fetch(), iPCref);  break;
    case 0x5c: iPCref = iPC;  FXLT0(Fetch(), iPCref);  break;
    case 0x5d: iPCref = iPC;  FL(0, Fetch(), iPCref);  break;
    case 0x5e: iPCref = iPC;  FXEQ0(Fetch(), iPCref);  break;
    case 0x6f: PRM(Fetch());     break;
    case 0x7f: PKXNE0(Fetch());  break;
    case 0x8f: PKGOTO(Fetch());  break;
    case 0x9f: PKXGE0(Fetch());  break;
    case 0xaf: PKGSB(Fetch());   break;
    case 0xbf: PKM(Fetch());     break;
    case 0xcf: PKXLT0(Fetch());  break;
    case 0xdf: PKRM(Fetch());    break;
    case 0xef: PKXEQ0(Fetch());  break;
    case 0xf1: c=Fetch();  PGOTO(c, Fetch());   break;
    case 0xf2: RTN(iPCref);      break;
    case 0xf3: c=Fetch();  PGSB(c, Fetch());    break;
    case 0xf4: c=Fetch();  PPM(c, Fetch());     break;
    case 0xf6: c=Fetch();  PPRM(c, Fetch());    break;
    case 0xf7: c=Fetch();  PXNE0(c, Fetch());   break;
    case 0xf8: c=Fetch();  PFL(2, c, Fetch());  break;
    case 0xf9: c=Fetch();  PXGE0(c, Fetch());   break;
    case 0xfa: c=Fetch();  PFL(3, c, Fetch());  break;
    case 0xfb: c=Fetch();  PFL(1, c, Fetch());  break;
    case 0xfc: c=Fetch();  PXLT0(c, Fetch());   break;
    case 0xfd: c=Fetch();  PFL(0, c, Fetch());  break;
    case 0xfe: c=Fetch();  PXEQ0(c, Fetch());   break;
    default:   Nope();
  }
  if (!flgSilent) State(Decomp(&iPC,0));
}


void Run(int flgSilent)
/* Клавиша С/П или команда :g запустила программу. */
{ iMode = 1;
  CommentLine("");
  do
  { Step(1);
    if (nBreaks)
    { int i; /* Останов в точках останова. */
      for (i=0; i<nBreaks; i++)
        if (iPC==iBreakAdr[i]) iMode = 0;
    }
  } while (iMode && (flgError==0));
  IntoAVT(0);
  if (!flgSilent) State(Decomp(&iPC,0));
}


