
/* display.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 5 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Графический экран и его функции.
 *
 * Функциональный узел ЭКВМ: индикатор.
 */

#include <stdlib.h>

#include "vk6.h"
#include "cpu.h"
#include "display.h"
#include "supp.h"
#include "reg.h"
#include "chargen.h"


unsigned char bScreen[1024];
int iX0 = 0, iY0 = 0, iGrAttr = 0, iGrDist = 0, iGrFont = 0, iGrInd = 0;


void DrawInit(void)
{ int i;
  iX0 = iY0 = 0;
  iGrAttr = 0;
  iGrDist = 0; /* Так в МК-161.  А зря, должно быть 1. */
  iGrFont = 0;
  iGrInd = 0;
  for (i=0; i<1024; i++) /* Забиваем мусором, как в МК-152. */
    bScreen[i] = (rand() & 255);
}


void Scroll(int c)
/* Прокрутка экрана, освободившиеся строки заполняются байтом c. */
{ int lines = Trunc(8), i;
  if (lines) /* При записи числа 0 действие не производится. */
  { if (lines<8)
      for (i=0; i<128*(8-lines); i++)
        bScreen[i] = bScreen[128*lines+i];
    for (i=128*(8-lines); i<1024; i++)
      bScreen[i] = c;
  }
}


void R9000w(void)
/* Программирование начальных координат объекта. */
{ iX0=Trunc(127);
  iY0=TruncY(63);
}


void R9000r(void)
/* Чтение начальных координат объекта. */
{ Push(iY0);
  Push(iX0);
}


void R9006w(void)
/* Вывод данных в графическую область памяти. */
{ bScreen[iGrInd++]=MkByte(RX);
  if (iGrInd>1023) iGrInd=0;
}


void R9006r(void)
/* Чтение данных из графической области памяти. */
{ Push(bScreen[iGrInd++]);
  if (iGrInd>1023) iGrInd=0;
}


void R9010w(void)
/* Инициализация графического экрана. */
{ int i = 0, f;
  switch (Trunc(3))
  { case 2: iGrDist = 1;  iX0=iY0=iGrFont=0;  iGrAttr = 0;
    case 0: f = 0;     break;
    case 3: iGrDist = 1;  iX0=iY0=iGrFont=0;  iGrAttr = 4;
    case 1: f = 0xff;  break;
  }
  while (i<1024) bScreen[i++] = f;
}


static int GetAdr(int *pMask, int x, int y)
{ *pMask = (1 << (y&7));
  return (y>>3)*128+x;
}


static void SetBits(int iAdr, int iMask)
{ int b = bScreen[iAdr], o = iMask, a = iGrAttr;
  if (a>3) o = 0,  a &= 3;
  switch (a)
  { case 1: o &= b;  break;
    case 2: o |= b;  break;
    case 3: o ^= b;  break;
  }
  o &= iMask;
  a = (~iMask)&b;
  bScreen[iAdr] = a|o;
}


static void OutBits(int iAdr, int iMask, int iBits)
{ int b = bScreen[iAdr], a = iGrAttr;
  if (a>3) iBits ^= iMask,  a &= 3;
  switch (a)
  { case 1: iBits &= b;  break;
    case 2: iBits |= b;  break;
    case 3: iBits ^= b;  break;
  }
  iBits &= iMask;
  a = (~iMask)&b;
  bScreen[iAdr] = a|iBits;
}


static void OutByte(int iAdr, int iByte)
{ int b = bScreen[iAdr], a = iGrAttr;
  if (a>3) iByte ^= 255,  a &= 3;
  switch (a)
  { case 0: bScreen[iAdr] = iByte;   break;
    case 1: bScreen[iAdr] &= iByte;  break;
    case 2: bScreen[iAdr] |= iByte;  break;
    case 3: bScreen[iAdr] ^= iByte;  break;
  }
}


static void BarEnd(int iAdr, int n, int line)
{ int iMask = (1<<(line+1))-1;
  int iO = iMask, a = iGrAttr;
  if (a>3) iO = 0,  a &= 3;
  while (n--)
  { int b = bScreen[iAdr], o = iO;
    switch (a)
    { case 1: o &= b;  break;
      case 2: o |= b;  break;
      case 3: o ^= b;  break;
    }
    o &= iMask;
    bScreen[iAdr++] = ( ((~iMask)&b)|o );
  }
}


static void BarStart(int iAdr, int n, int line)
{ int iMask = ((1<<(8-line))-1) << line;
  int iO = iMask, a = iGrAttr;
  if (a>3) iO = 0,  a &= 3;
  while (n--)
  { int b = bScreen[iAdr], o = iO;
    switch (a)
    { case 1: o &= b;  break;
      case 2: o |= b;  break;
      case 3: o ^= b;  break;
    }
    o &= iMask;
    bScreen[iAdr++] = ( ((~iMask)&b)|o );
  }
}


static void SetByte(int iAdr, int n)
{ switch (iGrAttr)
  { case 0:
    case 2:  while (n--) bScreen[iAdr++] = 255;  break;
    case 1:
    case 6:
    case 7:  break;
    case 3:  while (n--) bScreen[iAdr++] ^= 255;  break;
    case 4:
    case 5:  while (n--) bScreen[iAdr++] = 0;  break;
  }
}


static void Bar(int dx, int dy)
/* Вывод прямоугольника шириной dx и высотой dy. */
{ int a = (iY0>>3)*128+iX0, i,
      t = (iY0 & 7),  tk = (iY0+dy-1)&7;

  /* клиппинг */
  if ((iX0>63) || (iY0>127) || (dx==0) || (dy==0)) return;
  if (iX0+dx>128) dx = 128 - iX0;
  if (iY0+dy>64)  dy = 64 - iY0;

  if ( ((iY0+dy-1)|7) == (iY0|7) )
  { /* весь прямоугольник находится внутри одной строки */
    if (t==0) BarEnd(a, dx, tk);
    else if (tk==7) BarStart(a, dx, t);
    else /* есть точки вверху и внизу прямоугольника */
    { int mask = (((1 << (tk-t+1))-1) << t);
      for (i = 0; i<dx; i++) SetBits(a++, mask);
    }
  } else /* прямоугольник занимает несколько строк */
  { if (tk<7) { BarEnd( ((iY0+dy-1)>>3)*128+iX0, dx, tk);  dy -= tk+1; }
    if (t) { BarStart(a, dx, t);  a += 128;  dy -= 8-t; }
    while (dy>0)
    { SetByte(a, dx);
      a += 128;
      dy -= 8;
    }
  }
}


static void LineHor(int x, int y, int n)
/* Горизонтальная линия от (x, y) длиной n. */
{ int iMask;
  int iAdr = GetAdr(&iMask, x, y);
  while (n--)
    SetBits(iAdr++, iMask);
}


static void LineVert(int x, int y, int n)
/* Вертикальная линия от (x, y) длиной n. */
{ int a = (y>>3)*128+x, i,
      t = (y & 7),  tk = (y+n-1)&7;
  if ( ((y+n-1)|7) == (y|7) )
    SetBits(a, ((1 << (tk-t+1))-1) << t);
  else /* линия занимает несколько строк */
  { if (tk<7)
    { SetBits( ((y+n-1)>>3)*128+x, (1<<(tk+1))-1 );
      n -= tk+1;
    }
    if (t)
    { SetBits(a, ((1<<(8-t))-1) << t);
      a += 128;
      n -= 8-t;
    }
    while (n>0)
    { SetBits(a, 255);
      a += 128;
      n -= 8;
    }
  }
}


static void Frame(int dx, int dy)
/* Нарисовать рамку, если она вмещается. */
{ if ((dx>=3) && (dy>=3) && (iX0+dx<=128) && (iY0+dy<=64))
  { LineHor(iX0, iY0, dx-1);
    LineVert(iX0+dx-1, iY0, dy-1);
    LineHor(iX0+1, iY0+dy-1, dx-1);
    LineVert(iX0, iY0+1, dy-1);
  }
}


static void Line(int x1, int y1)
/* Нарисовать линию до (x1, y1) включительно.        */
/* Алгоритм Брезенхейма для восьмисвязной развёртки. */
{ int x = iX0, y = iY0;
  int dx = AbsI(x1-x), dy = AbsI(y1-y);
  int iDir = 1, i, m;
  if (dx>=dy)
  { if (x1<x) SwapI(&x, &x1), SwapI(&y, &y1);
    if (dy==0) LineHor(x, y, dx+1);
    else /* халява кончилась */
    { int a = GetAdr(&m, x, y), eps = -dx;
      if (y1<y) iDir = -1;
      for (i=0; i<=dx; i++)
      { SetBits(a++, m);
        eps += 2*dy;
        if (eps>=0)
        { eps -= 2*dx; /* Смещаемся по оси Oy. */
          switch (iDir)
          { case 1:  if (m<128) m<<=1;
                     else m=1, a+=128;
                     break;
            case -1: if (m>1) m>>=1;
                     else m=128, a-=128;
                     break;
          }
        } /* if eps */
      } /* for i */
    } /* if (dy==0) */
  } else
  { if (y1<y) SwapI(&x, &x1), SwapI(&y, &y1);
    if (dx==0) LineVert(x, y, dy+1);
    else /* вторая халява кончилась */
    { int a = GetAdr(&m, x, y), eps = -dy;
      if (x1<x) iDir = -1;
      for (i=0; i<=dy; i++)
      { SetBits(a, m);
        if (m<128) m<<=1; else m=1, a+=128;
        eps += 2*dx;
        if (eps>=0) eps -= 2*dy, a += iDir;
      } /* for i */
    } /* if (dx==0) */
  } /* if (dx>=dy) */
}


static void BitBltP(int p, int dx, int dy)
/* Из памяти программ на экран. */
{ int ln = (iY0>>3);
  int t = (iY0 & 7),  tk = (iY0+dy-1)&7;
  int a = ln*128+iX0,  m = (1<<(8-t))-1, i, b, b1;
  if (iX0+dx>128) dx = 128 - iX0;
  if (dy<=8)
  { int m1 = (1<<dy)-1;
    if ( (iY0|7)==((iY0+dy-1)|7) )
      for (i=0; i<dx; i++)
      { b = (bProg[p++]&m1);
        OutBits(a++, m1<<t, b<<t);
      }
    else /* вывод байта в две строки */
      for (i=0; i<dx; i++)
      { b = (bProg[p]&m);
        OutBits(a, m<<t, b<<t);
        b = (bProg[p++]&m1);
        if (ln<7) OutBits(a+128, m1>>(8-t), b>>(8-t));
        a++;
      }
  } else /* образ из несколько строк */
  { int m1 = (1<<(tk+1))-1;
    int n = (dy+t-tk-9)/8;
    int flgBottom = 1;
    if (ln+n>6) flgBottom = 0;
    else if (tk==7) n++, flgBottom = 0;
    if (ln+n>7) n = 7-ln;
    while (dx--)
    { int a1 = a++, p1 = p;
      p += (dy+7)/8;
      b = (bProg[p1++]&m);
      OutBits(a1, m<<t, b<<t);
      a1 += 128;
      for (i=0; i<n; i++)
      { b1 = bProg[p1++];
        b >>= 8-t;
        b |= (b1<<t);
        OutByte(a1, b & 255);
        a1 += 128;
        b = b1;
      }
      if (flgBottom)
      { if (tk>t) b1 = bProg[p1]; else b1 = 0;
        b >>= 8-t;
        b |= (b1<<t);
        OutBits(a1, m1, b&m1);
      }
    }
  }
}


static void BitBltR(int r, int dx, int dy)
{ int ln = (iY0>>3);
  int t = (iY0 & 7), tk = (iY0+dy-1)&7;
  int a = ln*128+iX0, m = (1<<(8-t))-1, i, b, b1;
  if (iX0+dx>128) dx = 128 - iX0;
  if (dy<=8)
  { int m1 = (1<<dy)-1;
    if ( (iY0|7)==((iY0+dy-1)|7) )
      for (i=0; i<dx; i++)
      { b = (bR[r++]&m1);
        OutBits(a++, m1<<t, b<<t);
      }
    else /* вывод байта в две строки */
      for (i=0; i<dx; i++)
      { b = (bR[r]&m);
        OutBits(a, m<<t, b<<t);
        b = (bR[r++]&m1);
        if (ln<7) OutBits(a+128, m1>>(8-t), b>>(8-t));
        a++;
      }
  } else /* образ из несколько строк */
  { int m1 = (1<<(tk+1))-1;
    int n = (dy+t-tk-9)/8;
    int flgBottom = 1;
    if (ln+n>6) flgBottom = 0;
    else if (tk==7) n++, flgBottom = 0;
    if (ln+n>7) n = 7-ln;
    while (dx--)
    { int a1 = a++, r1 = r;
      r += (dy+7)/8;
      b = (bR[r1++]&m);
      OutBits(a1, m<<t, b<<t);
      a1 += 128;
      for (i=0; i<n; i++)
      { b1 = bR[r1++];
        b >>= 8-t;
        b |= (b1<<t);
        OutByte(a1, b & 255);
        a1 += 128;
        b = b1;
      }
      if (flgBottom)
      { if (tk>t) b1 = bR[r1]; else b1 = 0;
        b >>= 8-t;
        b |= (b1<<t);
        OutBits(a1, m1, b&m1);
      }
    }
  }
}


static void DrawChar(TCharacter *pCh, int dx, int dy)
{ int ln = (iY0>>3), p = 0;
  int t = (iY0 & 7), tk = (iY0+dy-1)&7;
  int a = ln*128+iX0, m = (1<<(8-t))-1, i, b, b1;
  if (iX0+dx>128) dx = 128 - iX0;
  if (dy<=8)
  { int m1 = (1<<dy)-1;
    if ( (iY0|7)==((iY0+dy-1)|7) )
      for (i=0; i<dx; i++)
      { b = ((*pCh)[p++]&m1);
        OutBits(a++, m1<<t, b<<t);
      }
    else /* вывод байта в две строки */
      for (i=0; i<dx; i++)
      { b = ((*pCh)[p]&m);
        OutBits(a, m<<t, b<<t);
        b = ((*pCh)[p++]&m1);
        if (ln<7) OutBits(a+128, m1>>(8-t), b>>(8-t));
        a++;
      }
  } else /* образ из несколько строк */
  { int m1 = (1<<(tk+1))-1;
    int n = (dy+t-tk-9)/8;
    int flgBottom = 1;
    if (ln+n>6) flgBottom = 0;
    else if (tk==7) n++, flgBottom = 0;
    if (ln+n>7) n = 7-ln;
    while (dx--)
    { int a1 = a++, p1 = p;
      p += (dy+7)/8;
      b = ((*pCh)[p1++]&m);
      OutBits(a1, m<<t, b<<t);
      a1 += 128;
      for (i=0; i<n; i++)
      { b1 = (*pCh)[p1++];
        b >>= 8-t;
        b |= (b1<<t);
        OutByte(a1, b & 255);
        a1 += 128;
        b = b1;
      }
      if (flgBottom)
      { if (tk>t) b1 = (*pCh)[p1]; else b1 = 0;
        b >>= 8-t;
        b |= (b1<<t);
        OutBits(a1, m1, b&m1);
      }
    }
  }
}


void PrintChar(unsigned char c)
/* Вывести (очередной) символ на граф. экран. */
{ int dx = tblFontSize[iGrFont][c], dy = tblFontHeight[iGrFont];
  if ((iX0<128) && dx)
  { DrawChar(tblFont[iGrFont][c], dx, dy);
    iX0 += dx;
    if (iX0<128)
    { iGrAttr ^= 4;
      Bar(iGrDist, dy);
      iGrAttr ^= 4;
      iX0 += iGrDist;
    }
  }
}


void R9011w(void)
/* Вывод точки. */
{ int x=Trunc(127), y=TruncY(63), iMask;
  int iAdr = GetAdr(&iMask, x, y);
  SetBits(iAdr, iMask); 
}


void R9011r(void)
/* Чтение состояния точки на графическом экране. */
{ int x = Trunc(127), y = TruncY(63), iMask;
  int iAdr = GetAdr(&iMask, x, y);
  if (bScreen[iAdr] & iMask) Push(1); else Push(0);
}


void R9012w(void)
/* Вывод линии. */
{ int x = MkByte(RX), y = MkByte(RY);
  if ((RX>=0) && (RY>=0) && (x<128) && (y<64) && (iX0<128))
    Line(x, y);
}


void R9013w(void)
/* Вывод прямоугольника. */
{ int dx = Trunc(128), dy = TruncY(64);
  if (iX0<128) Bar(dx, dy);
}


void R9014w(void)
/* Вывод рамки. */
{ int dx = MkByte(RX), dy = MkByte(RY);
  if (iX0<128) Frame(dx, dy);
}


void R9015w(void)
/* Вывод графического образа из памяти программ. */
{ int a = Trunc(9999);
  if ((iX0<128) && (a<9998))
  { int dx = bProg[a++];
    int dy = bProg[a++];
    if ((dx>0) && (dy>0) && (a+(dy+7)/8*dx<=10000))
      BitBltP(a, dx, dy);
  }
}


void R9016w(void)
/* Вывод графического образа из памяти данных. */
{ int r = Trunc(8167);
  if ((iX0<128) && (r>=1000) && (r<8166))
  { int dx = bR[r++];
    int dy = bR[r++];
    if ((dx>0) && (dy>0) && (r+(dy+7)/8*dx<=8168))
      BitBltR(r, dx, dy);
  }
}


void R9020w(void)
/* Вывод символа. */
{ int c = MkByte(RX);
  PrintChar(c);
}


void R9021w(void)
/* Вывод строки символов из памяти программ. */
{ int a = Trunc(9999);
  while (a<=9999)
  { int c = bProg[a++];
    if ((c==0) || (c==255)) return;
    PrintChar(c);
  }
}


void R9022w(void)
/* Вывод строки символов из памяти данных. */
{ int a = Trunc(8167) - 1000;
  if (a>=0)
    while (a<=7167)
    { int c = bR[a++];
      if ((c==0) || (c==255)) return;
      PrintChar(c);
    }
}


void R9023w(void)
/* Вывод числа с естественной или плавающей запятой. */
{ unsigned char *s = NumToStr(RX);
  while (*s)
    PrintChar(*s++);
}


void R9024w(void)
/* Вывод числа в форматированном виде. */
{ int y = Trun1Y(8), z = TruncZ(8);
  int c;
  unsigned char *s = NumToStrFix(RX, y, z);
  while (c=*s++)
    if ((iGrFont==0) && (c==' ')) PrintChar(31);
    else PrintChar(c);
}


