
/* keyboard.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 21 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Клавиатура ВК-6: что происходит при нажатии на клавиши.
 *
 * Функциональный узел ЭКВМ: клавиатура.
 */

#include "cpu.h"
#include "reg.h"
#include "decomp.h"
#include "keyboard.h"
#include "clock.h"
#include "comment.h"
#include "supp.h"


int iKbdState = 0;
/* 0 основное значение клавиши
   1 F
   2 K
   3 P
   4 PF: L0, L1, L2, L3
   5 РК
   6 РР
   7 требуется регистр: П, ИП, КП, КИП, КБП, КПП, Кx<0, Кx=0, Кx>=0, Кx!=0
   8 требуется байт: БП, ПП, FL0, FL1, FL2, FL3, РП, РИП, PKП, PKИП, РКБП, PKПП,
                     Fx<0, Fx=0, Fx>=0, Fx!=0, PKx<0, PKx=0, PKx>=0, PKx!=0
   9 требуются два байта: РБП, РПП, РРП, РРИП, Рx<0, Px=0, Px>=0, Px!=0,
                          PFL0, PFL1, PFL2, PFL3
 */

static int iKbdCOP = 0; /* Код ожидаемой операции */
static int iBCD = 0;    /* Введённое число        */
int iNumCnt = 0;        /* Кол-во введённых цифр  */

static int iKbdHead = 0, iKbdTail = 0, iButton = 255;
static unsigned char sKbdBuffer[8];


static void KbdPut(unsigned char c)
/* Разместить клавишу в буфере клавиатуры. */
{ int q;
  sKbdBuffer[iKbdHead] = c;
  q = iKbdHead - iKbdTail;
  if (q<0) q+=8;
  if (q<4)
    if (iKbdHead==7) iKbdHead = 0;
    else iKbdHead++;
}


unsigned char KbdGet(void)
/* Считать клавишу из буфера клавиатуры. */
{ if (iKbdHead!=iKbdTail)
  { unsigned char c = sKbdBuffer[iKbdTail];
    if (iKbdTail==7) iKbdTail = 0;
    else iKbdTail++;
    return c;
  } else return 255;
}


void KbdReset(void)
/* Сбросить буфер клавиатуры. */
{ iKbdTail = iKbdHead;
  iButton = 255;
}


void R9028r(void)
/* Чтение состояния клавиатуры. */
{ Push(iButton);
}


void IntoFPRG()
/* F ПРГ, переход в режим программирования. */
{ iMode = 2;
  iKbdState = 0;
  CommentLine("F \x8f\x90\x83"); /* F ПРГ */
}


void IntoRun()
/* Владелец нажал С/П и запустил программу. */
{ iMode = 1;
  KbdReset();
}


void IntoPPRG(void)
/* Р ПРГ, переход в режим ввода байт. */
{ iMode = 3;
  iKbdState = 8;
  iNumCnt = 0;
  CommentLine(flgLat? "P PRG": "\x90 \x8f\x90\x83"); /* Р ПРГ */
}


static void Compile(int c)
/* Записать одношаговую команду в память программ. */
{ bProg[iPC++] = c;
  if (iPC>9999) iPC = 0;
  if (iMode==2) State1(c);
}


static void Compile2(int c, int bcd)
/* Записать двухшаговую команду в память программ. */
{ bProg[iPC++] = c;
  if (iPC>9999) iPC = 0;
  bProg[iPC++] = bcd;
  if (iPC>9999) iPC = 0;
}


static void Compile3(int c, int bcd1, int bcd2)
/* Записать трёхшаговую команду в память программ. */
{ bProg[iPC++] = c;
  if (iPC>9999) iPC = 0;
  bProg[iPC++] = bcd1;
  if (iPC>9999) iPC = 0;
  bProg[iPC++] = bcd2;
  if (iPC>9999) iPC = 0;
}


static void F(void)
/* Нажатие клавиши F */
{ CommentLine("F");
  iKbdState = 1;
}


static void K(void)
/* Нажатие клавиши К */
{ CommentLine(flgLat? "K" : "\x8a");
  iKbdState = 2;
}


static void P(void)
/* Нажатие клавиши Р */
{ CommentLine(flgLat? "P" : "\x90");
  iKbdState = 3;
}


static void FCF(void)
/* Комбинация F CF */
{ CommentLine("F CF");
}


static void ModePPRG(int c)
/* Клавиши с префиксом в режиме Р ПРГ */
{ switch (iKbdState)
  { case 1: /* F */
      switch (c)
      { case KEY_NEG:
          IntoAVT(0);
          CommentLine(flgLat? "F AVT" : "F \x80\x82\x92"); /* F АВТ */
          iKbdState = 0;
          break;
        case KEY_EE:
          IntoFPRG();
          break;
        case KEY_CX:
          FCF();
          iKbdState = 8;
          break;
      }
      break;
    case 3: /* P */
      switch (c)
      { case KEY_F:
          F();
          break;
        case KEY_NEG:
          IntoAVT(0);
          CommentLine(flgLat? "P AVT" : "\x90 \x80\x82\x92"); /* Р АВТ */
          iKbdState = 0;
          break;
        case KEY_EE:
          IntoPPRG();
          break;
        case KEY_CX:
          CommentLine(flgLat? "P CF" : "\x90 CF"); /* P CF */
          iKbdState = 8;
          iBCD = 0;
          break;
      }
      break;
  }
}


static void Key(int c)
/* Нажата клавиша, без префиксов. */
{ if (c<=KEY_9) { if (iMode==2) Compile(c); else Digit(c); }
  else switch (c)
  { case KEY_TOCHKA:
      if (iMode==2) Compile(0x0a); else Tochka();
      break;
    case KEY_NEG:
      if (iMode==2) Compile(0x0b); else Negate();
      break;
    case KEY_EE:
      if (iMode==2) Compile(0x0c); else EE();
      break;
    case KEY_CX:
      if (iMode==2) Compile(0x0d); else Cx();
      break;
    case KEY_ENTER:
      if (iMode==2) Compile(0x0e); else Enter();
      break;
    case KEY_PLUS:
      if (iMode==2) Compile(0x10); else Plus();
      break;
    case KEY_MINUS:
      if (iMode==2) Compile(0x11); else Minus();
      break;
    case KEY_MUL:
      if (iMode==2) Compile(0x12); else Multiply();
      break;
    case KEY_DIV:
      if (iMode==2) Compile(0x13); else Divide();
      break;
    case KEY_XCHG:
      if (iMode==2) Compile(0x14); else XCHG();
      break;
    case KEY_F: F(); break;
    case KEY_K: K(); break;
    case KEY_P: P(); break;
    case KEY_RET:
      if (iMode==2) Compile(0x52); else B0();
      break;
    case KEY_RS:
      if (iMode==2) Compile(0x50); else IntoRun();
      break;
    case KEY_RM:
      CommentLine(flgLat? "RM ":"\x88\x8f "); /* ИП */
      iKbdState = 7;
      iBCD = 0;
      iKbdCOP = 0x60;
      break;
    case KEY_M:
      CommentLine(flgLat? "M ":"\x8f "); /* П */
      iKbdState = 7;
      iBCD = 0;
      iKbdCOP = 0x40;
      break;
    case KEY_GOTO:
      CommentLine(flgLat? "GOTO ":"\x81\x8f "); /* БП */
      iKbdState = 8;  iBCD = iNumCnt = 0;
      iKbdCOP = 0x51;
      break;
    case KEY_GSB:
      if (iMode==2)
      { CommentLine(flgLat? "GSB ":"\x8f\x8f "); /* ПП */
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x51;
      } else Step(0);
      break;
  }
}


static void KeyF(int c)
/* Нажата клавиша, после F */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_0:
      if (iMode==2) Compile(0x15); else F10X();
      break;
    case KEY_1:
      if (iMode==2) Compile(0x16); else FEXP();
      break;
    case KEY_2:
      if (iMode==2) Compile(0x17); else Flg();
      break;
    case KEY_3:
      if (iMode==2) Compile(0x18); else Fln();
      break;
    case KEY_4:
      if (iMode==2) Compile(0x19); else Farcsin();
      break;
    case KEY_5:
      if (iMode==2) Compile(0x1a); else Farccos();
      break;
    case KEY_6:
      if (iMode==2) Compile(0x1b); else Farctg();
      break;
    case KEY_7:
      if (iMode==2) Compile(0x1c); else Fsin();
      break;
    case KEY_8:
      if (iMode==2) Compile(0x1d); else Fcos();
      break;
    case KEY_9:
      if (iMode==2) Compile(0x1e); else Ftg();
      break;
    case KEY_TOCHKA:
      if (iMode==2) Compile(0x25); else FR();
      break;
    case KEY_NEG:
      CommentLine(flgLat? "F AVT" : "F \x80\x82\x92"); /* F АВТ */
      IntoAVT(0);
      break;
    case KEY_EE:
      IntoFPRG();
      break;
    case KEY_CX:
      FCF();
      break;
    case KEY_ENTER:
      if (iMode==2) Compile(0x0f); else FANS();
      break;
    case KEY_PLUS:
      if (iMode==2) Compile(0x20); else FPI();
      break;
    case KEY_MINUS:
      if (iMode==2) Compile(0x21); else FSQRT();
      break;
    case KEY_MUL:
      if (iMode==2) Compile(0x22); else FX2();
      break;
    case KEY_DIV:
      if (iMode==2) Compile(0x23); else F1X();
      break;
    case KEY_XCHG:
      if (iMode==3) Compile(0x24); else FXY();
      break;
    case KEY_SFWD: /* Fx<0  */
      if (iMode==2)
      { CommentLine(flgLat? "F X<0 ":"F x<0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x5c;
      }
      break;
    case KEY_SBCK: /* Fx=0  */
      if (iMode==2)
      { CommentLine(flgLat? "F X=0 ":"F x=0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x5e;
      }
      break;
    case KEY_RET:  /* Fx>=0 */
      if (iMode==2)
      { CommentLine(flgLat? "F X>=0 ":"F x\xb1" "0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x59;
      }
      break;
    case KEY_RS:   /* Fx!=0 */
      if (iMode==2)
      { CommentLine(flgLat? "F X!=0 ":"F x\xb2" "0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x57;
      }
      break;
    case KEY_RM:   /* FL0   */
      if (iMode==2)
      { CommentLine("F L0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x5d;
      }
      break;
    case KEY_M:    /* FL1   */
      if (iMode==2)
      { CommentLine("F L1 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x5b;
      }
      break;
    case KEY_GOTO: /* FL2   */
      if (iMode==2)
      { CommentLine("F L2 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x58;
      }
      break;
    case KEY_GSB:  /* FL3   */
      if (iMode==2)
      { CommentLine("F L3 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x5a;
      }
      break;
  }
}


static void KeyK(int c)
/* Нажата клавиша, после К */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_0:
      if (iMode==2) Compile(0x54); else KNOP();
      break;
    case KEY_1:
      if (iMode==2) Compile(0x55); else KSCR();
      break;
    case KEY_2:
      if (iMode==2) Compile(0x56); else KGRPH();
      break;
    case KEY_3:
      if (iMode==2) Compile(0x30); else KDMS();
      break;
    case KEY_4:
      if (iMode==2) Compile(0x31); else KABS();
      break;
    case KEY_5:
      if (iMode==2) Compile(0x32); else KSGN();
      break;
    case KEY_6:
      if (iMode==2) Compile(0x33); else KDM();
      break;
    case KEY_7:
      if (iMode==2) Compile(0x34); else KINT();
      break;
    case KEY_8:
      if (iMode==2) Compile(0x35); else KFRAC();
      break;
    case KEY_9:
      if (iMode==2) Compile(0x36); else Kmax();
      break;
    case KEY_TOCHKA:
      if (iMode==2) Compile(0x37); else KAND();
      break;
    case KEY_NEG:
      if (iMode==2) Compile(0x38); else KOR();
      break;
    case KEY_EE:
      if (iMode==2) Compile(0x39); else KXOR();
      break;
    case KEY_CX:
      if (iMode==2) Compile(0x3a); else KNOT();
      break;
    case KEY_ENTER:
      if (iMode==2) Compile(0x3b); else KRAN();
      break;
    case KEY_PLUS:
      if (iMode==2) Compile(0x26); else KMD();
      break;
    case KEY_MINUS:
      if (iMode==2) Compile(0x27); else KMINUS();
      break;
    case KEY_MUL:
      if (iMode==2) Compile(0x28); else KPRGM();
      break;
    case KEY_DIV:
      if (iMode==2) Compile(0x29); else KDIV();
      break;
    case KEY_XCHG:
      if (iMode==2) Compile(0x2a); else KMSD();
      break;
    case KEY_F:
      F();
      break;
    case KEY_SFWD: /* Kx<0  */
      if (iMode==2)
      { CommentLine(flgLat? "K X<0 ":"\x8a x<0 ");
        iKbdState = 7;
        iBCD = 0;
        iKbdCOP = 0xc0;
      }
      break;
    case KEY_SBCK: /* Kx=0  */
      if (iMode==2)
      { CommentLine(flgLat? "K X=0 ":"\x8a x=0 ");
        iKbdState = 7;
        iBCD = 0;
        iKbdCOP = 0xe0;
      }
      break;
    case KEY_RET:  /* Kx>=0 */
      if (iMode==2)
      { CommentLine(flgLat? "K X>=0 ":"\x8a x\xb1" "0 ");
        iKbdState = 7;
        iBCD = 0;
        iKbdCOP = 0x90;
      }
      break;
    case KEY_RS:   /* Kx!=0 */
      if (iMode==2)
      { CommentLine(flgLat? "K X!=0 ":"\x8a x\xb2" "0 ");
        iKbdState = 7;
        iBCD = 0;
        iKbdCOP = 0x70;
      }
      break;
    case KEY_RM:
      CommentLine(flgLat? "K RM ":"\x8a \x88\x8f "); /* ИП */
      iKbdState = 7;
      iBCD = 0;
      iKbdCOP = 0xd0;
      break;
    case KEY_M:
      CommentLine(flgLat? "K M ":"\x8a \x8f "); /* П */
      iKbdState = 7;
      iBCD = 0;
      iKbdCOP = 0xb0;
      break;
    case KEY_GOTO:
      CommentLine(flgLat? "K GOTO ":"\x8a \x81\x8f "); /* БП */
      iKbdState = 7;
      iBCD = 0;
      iKbdCOP = 0x80;
      break;
    case KEY_GSB:
      if (iMode==2)
      { CommentLine(flgLat? "K GSB ":"\x8a \x8f\x8f "); /* ПП */
        iKbdState = 7;
        iBCD = 0;
        iKbdCOP = 0xa0;
      }
      break;
  }
}


static void KeyP(int c)
/* Нажата клавиша, после Р */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_NEG:
      CommentLine(flgLat? "P AVT" : "\x90 \x80\x82\x92"); /* Р АВТ */
      IntoAVT(0);
      break;
    case KEY_EE:
      IntoPPRG();
      break;
    case KEY_CX:
      CommentLine(flgLat? "P CF": "\x90 \x91" "F");
      break;
    case KEY_F:
      CommentLine(flgLat? "PF": "\x90" "F");
      iKbdState = 4;
      break;
    case KEY_K:
      CommentLine(flgLat? "PK": "\x90\x8a");
      iKbdState = 5;
      break;
    case KEY_P:
      CommentLine(flgLat? "PP": "\x90\x90");
      iKbdState = 6;
      break;
    case KEY_SFWD:
      if (iMode==2)
      { CommentLine(flgLat? "P X<0 ":"\x90 x<0 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xfc;
      }
      break;
    case KEY_SBCK:
      if (iMode==2)
      { CommentLine(flgLat? "P X=0 ":"\x90 x=0 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xfe;
      }
      break;
    case KEY_RET:
      if (iMode==2)
      { CommentLine(flgLat? "P X>=0 ":"\x90 x\xb1" "0 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xf9;
      }
      break;
    case KEY_RS:
      if (iMode==2)
      { CommentLine(flgLat? "P X!=0 ":"\x90 x\xb2" "0 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xf7;
      }
      break;
    case KEY_RM:
      CommentLine(flgLat? "P RM ":"\x90 \x88\x8f");
      iKbdState = 8;  iBCD = iNumCnt = 0;
      iKbdCOP = 0x6f;
      break;
    case KEY_M:
      CommentLine(flgLat? "P M ":"\x90 \x8f");
      iKbdState = 8;  iBCD = iNumCnt = 0;
      iKbdCOP = 0x4f;
      break;
    case KEY_GOTO:
      CommentLine(flgLat? "P GOTO ":"\x90 \x81\x8f");
      iKbdState = 9;  iBCD = iNumCnt = 0;
      iKbdCOP = 0xf1;
      break;
    case KEY_GSB:
      if (iMode==2)
      { CommentLine(flgLat? "P GSB ":"\x90 \x8f\x8f ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xf3;
      }
      break;
  }
}


static void KeyPF(int c)
/* Нажата клавиша, после PF */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_CX:
      CommentLine(flgLat? "PF CF": "\x90" "F CF");
      break;
    case KEY_F:
      F();
      break;
    case KEY_RM:
      if (iMode==2)
      { CommentLine(flgLat? "PF L0 ":"\x90" "F L0 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xfd;
      }
      break;
    case KEY_M:
      if (iMode==2)
      { CommentLine(flgLat? "PF L1 ":"\x90" "F L1 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xfb;
      }
      break;
    case KEY_RET:
      if (iMode==2)
      { CommentLine(flgLat? "PF L2 ":"\x90" "F L2 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xf8;
      }
      break;
    case KEY_RS:
      if (iMode==2)
      { CommentLine(flgLat? "PF L3 ":"\x90" "F L3 ");
        iKbdState = 9;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xfa;
      }
      break;
  }
}


static void KeyPK(int c)
/* Нажата клавиша, после PK */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_CX:
      CommentLine(flgLat? "PK CF": "\x90\x8a CF");
      break;
    case KEY_F:
      F();
      break;
    case KEY_SFWD:
      if (iMode==2)
      { CommentLine(flgLat? "PK X<0 ":"\x90\x8a x<0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xcf;
      }
      break;
    case KEY_SBCK:
      if (iMode==2)
      { CommentLine(flgLat? "PK X=0 ":"\x90\x8a x=0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0xef;
      }
      break;
    case KEY_RET:
      if (iMode==2)
      { CommentLine(flgLat? "PK X>=0 ":"\x90\x8a x\xb1" "0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x9f;
      }
      break;
    case KEY_RS:
      if (iMode==2)
      { CommentLine(flgLat? "PK X!=0 ":"\x90\x8a x\xb2" "0 ");
        iKbdState = 8;  iBCD = iNumCnt = 0;
        iKbdCOP = 0x7f;
      }
      break;
  }
}


static void KeyPP(int c)
/* Нажата клавиша, после PP */
{ iKbdState = 0;
  CommentLine("");
  switch (c)
  { case KEY_CX:
      CommentLine(flgLat? "PP CF": "\x90\x90 CF");
      break;
    case KEY_F:
      F();
      break;
    case KEY_RET:
      if (iMode==2)
      { CommentLine(flgLat? "PP RTN": "\x90\x90 \x82/\x8e");
        Compile(0xf2);
      }
      break;
    case KEY_RM:
      CommentLine(flgLat? "PP RM ":"\x90\x90 \x88\x8f");
      iKbdState = 9;  iBCD = iNumCnt = 0;
      iKbdCOP = 0xf6;
      break;
    case KEY_M:
      CommentLine(flgLat? "PP M ":"\x90\x90 \x8f");
      iKbdState = 9;  iBCD = iNumCnt = 0;
      iKbdCOP = 0xf4;
      break;
  }
}


static void DoReg(int c)
/* Завершён ввод регистровой команды. */
{ iKbdState = 0;
  if (iMode==2) Compile(iKbdCOP+c); /* Режим F ПРГ */
  else switch (iKbdCOP) /* Режим F АВТ поддерживает не все рег. команды */
  { case 0x40: M(c);      break;
    case 0x60: RM(c);     break;
    case 0x80: KGOTO(c);  break;
    case 0xb0: KM(c);     break;
    case 0xd0: KRM(c);    break;
  }
}


static void DoNN(int bcd)
/* Завершён ввод двухбайтовой команды. */
{ if (iMode==3) { Compile(bcd);  return; } /* Режим Р ПРГ */
  iKbdState = 0;
  if (iMode==2) Compile2(iKbdCOP, bcd); /* Режим F ПРГ */
  else switch (iKbdCOP) /* Режим F АВТ поддерживает не все 2-ш. команды */
  { case 0x4f: PM(bcd);        break;
    case 0x51: GOTO(bcd, iPC); break;
    case 0x6f: PRM(bcd);       break;
    case 0x8f: PKGOTO(bcd);    break;
    case 0xaf: PKGSB(bcd);     break;
    case 0xbf: PKM(bcd);       break;
    case 0xdf: PKRM(bcd);      break;
  }
}


static void DoNNNN(int bcd1, int bcd2)
/* Завершён ввод трёхбайтовой команды. */
{ iKbdState = 0;
  if (iMode==2) Compile3(iKbdCOP, bcd1, bcd2);
  else switch (iKbdCOP) /* Режим F АВТ поддерживает не все 3-ш. команды */
  { case 0xf1: PGOTO(bcd1, bcd2); break;
    case 0xf4: PPM(bcd1, bcd2);   break;
    case 0xf6: PPRM(bcd1, bcd2);  break;
  }
}


void ButtonRelease(int c)
/* Клавиша отпущена и больше не нажата. */
{ if (c==iButton) iButton = 255;
}


void ButtonPress(int c)
/* Обработка нажатия клавиши на калькуляторе ВК-6. */
{ if ((c<0) || (c>MAX_KEY)) return;
  iButton = c;
  /* На самом деле ВВОД, ВЫВОД и ВЫБОР не сбрасывают flgError. */
  flgError = 0;
  KbdClick();
  if (iMode==1) /* Режим выполнения программы */
  { if (c==KEY_RS) IntoAVT(0); else KbdPut(c);
  } else if (c==KEY_GRAD) switch (iMode)
  { case 3: /* РУС/ЛАТ */
    case 2: flgLat ^= 1;  break;
    case 0: GRAD();       break;
  } else if (iKbdState>=7) /* Происходит ввод числа */
  { if (c==KEY_F) { F();  return; } /* Для F CF  */
    if (c==KEY_P) { P();  return; } /* P ПРГ/АВТ */
    if ((iMode==3) && (iNumCnt==0)) /* В режиме Р ПРГ порой работают ШГ. */
    { if ((c==KEY_SFWD) || (c==KEY_DN))
      { if (iPC<9999) iPC++; else iPC = 0;
        return;
      } else if ((c==KEY_SBCK) || (c==KEY_UP))
      { if (iPC) iPC--; else iPC = 9999;
        return;
      } else if (c==KEY_RT)
      { iPC += 5;
        if (iPC>9999) iPC -= 10000;
        return;
      } else if (c==KEY_LT)
      { iPC -= 5;
        if (iPC<0) iPC += 10000;
        return;
      }
    }
    if ((c==KEY_MUL) && (iKbdState!=7)) c = 15; else if (c>KEY_ENTER) return;
    if (iKbdState==7) { DoReg(c); return; }
    if ((c>KEY_9) && (iMode!=3)) return;
    if ((iMode==3) && (iNumCnt==0)) CommentLine("");
    CommentStoreCh(strHex[c]);
    iBCD = iBCD*16 + c;
    iNumCnt++;
    if ((iNumCnt>3) && (iKbdState==9)) { DoNNNN(iBCD>>8, iBCD&255);  iNumCnt=0; }
    else if ((iNumCnt>1) && (iKbdState==8)) { DoNN(iBCD);  iNumCnt=0; }
  } else if ( (iKbdState==0) &&
              ( (c==KEY_SFWD) || ((iMode==2) && (c==KEY_DN)) ) )
  { if (iPC<9999) iPC++; else iPC = 0;
    if (c==KEY_SFWD) CommentLine("\x98\x83\xc5"); /* ШГ-> */
  } else if ( (iKbdState==0) &&
              ( (c==KEY_SBCK) || ((iMode==2) && (c==KEY_UP)) ) )
  { if (iPC) iPC--; else iPC = 9999;
    if (c==KEY_SBCK) CommentLine("\x98\x83\xc4"); /* ШГ<- */
  } else if ((iKbdState==0) && (c==KEY_RT) && (iMode>=2))
  { iPC += 5;
    if (iPC>9999) iPC -= 10000;
  } else if ((iKbdState==0) && (c==KEY_LT) && (iMode>=2))
  { iPC -= 5;
    if (iPC<0) iPC += 10000;
  } else if (iMode==3) ModePPRG(c); /* P ПРГ: режим ввода байтов */
  else switch (iKbdState)         /* Режимы 0 и 2, F АВТ и F ПРГ */
  { case 0: Key(c);    break;
    case 1: KeyF(c);   break;
    case 2: KeyK(c);   break;
    case 3: KeyP(c);   break;
    case 4: KeyPF(c);  break;
    case 5: KeyPK(c);  break;
    case 6: KeyPP(c);  break;
  }
}

