
/* mkchargen.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 5 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Компилятор свободного знакогенератора ВК-6.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cpu.h"


unsigned char tblFontHeight[3] = { 8, 14, 19 };
unsigned char tblFontSize[3][256];
unsigned char tblCharSize[3][256];
unsigned char *tblFont[3][256];

int n;
int flgVerbose;
char s[100];

char *tblPreface[] =
{ "\n",
  "/* chargen.c\n",
  " * ВК-6 " STR_VERSION "\n",
  " * (c) Васильев И.В., Москва, 5 марта 2010 г.\n",
  " *\n",
  " * Программа распространяется под лицензией GNU GPL v3.\n",
  " *\n",
  " * Знакогенератор ВК-6.\n",
  " * Кодировка -- вариант 866, совместимый с МК-161.\n",
  " *\n",
  " * АВТОМАТИЧЕСКИ СОЗДАННЫЙ ФАЙЛ, НЕ ИЗМЕНЯТЬ!\n",
  " */\n",
  "\n",
  "\n",
  "#include \"chargen.h\"\n",
  "\n",
  "unsigned char tblFontHeight[3] = { 8, 14, 19 };\n",
  "\n",
  "unsigned char tblFontSize [3][256] =\n",
  "{",
  0
};


void InitTables(void)
{ int i, j;
  for (i=0; i<3; i++)
    for (j=0; j<255; j++)
      tblFontSize[i][j] = 0,
      tblCharSize[i][j] = 0,
      tblFont[i][j] = NULL;
}


int ReadChar(FILE *f, int c, int fnt)
/* Считать символ, вернуть 0 при ошибке. */
{ int dy = tblFontHeight[fnt], dx = -1, m = 1, ct = 0, i, j, q;
  static unsigned char tbl[4][25], *pc;
  if (flgVerbose)
    printf("Считываю символ '%c': код %X, шрифт %X.\n", c<' '?'.':c, c, fnt);
  for (i=0; i<4; i++)
    for (j=0; j<25; j++)
      tbl[i][j]=0;
  for (i=0; i<dy; i++)
  { j = q = 0;
    fgets(s, 100, f);
    n++;
    while (s[j])
      switch (s[j++])
      { case 'x':
          tbl[ct][q] |= m;
        case '.':
          if (++q>=25)
          { printf("Строка %d слишком длинна.\n", n);
            return 0;
          }
          break;
        case '\n':
        case '\t':
        case ' ':
          break;
        default:
          printf("Неверное изображение в строке %d.\n", n);
          return 0;
      }
    if (m==128) m=1, ct++;
    else m<<=1;
    if (dx==-1) dx = q;
    if (q!=dx)
    { printf("Ширина символа %02x в строке %d должна быть %d.\n",
        c, n, dx);
      return 0;
    }
  }
  if (ct && (m==1)) ct--;
  tblCharSize[fnt][c] = m = (ct+1)*dx;
/*  printf("Символ занял %d столбцов по %d байт.\n", dx, ct+1); */
  if (tblFont[fnt][c] != NULL)
  { printf("Повторное определение символа %02x в строке %d.\n",
      c, n);
    return 0;
  }
  tblFont[fnt][c] = pc = malloc(m);
  if (pc==NULL)
  { puts("Не хватило памяти.");
    return 0;
  }
  for (j=0; j<dx; j++)
    for (i=0; i<=ct; i++)
      *pc++ = tbl[i][j];
  tblFontSize[fnt][c] = dx;
  return 1;
}


void DumpChar(FILE *f, int c, int fnt)
/* Записать символ и освободить память. */
{ unsigned char *pc = tblFont[fnt][c];
  if (pc!=NULL)
  { int i = 0, n = tblCharSize[fnt][c];
    while (i<n)
    { fprintf(f, "0x%02x", pc[i++]);
      if (i!=n)
        if ((i&7)==0) fputs(",\n  ", f);
        else fputs(", ", f);
    }
    free(pc);
  }
}


int Read(FILE *f)
/* Считать файл, вернуть 0 при ошибке. */
{ int c, fnt;
  n = 0;
  for (;;)
  { fgets(s, 100, f);
    n++;
    switch (s[0])
    { case '-':
        return 1;
      case ':':
        sscanf(s, ":%x,%d", &c, &fnt);
        if ((c>255) || (fnt>2))
          printf("Неверный символ в строке %d.\n", n);
        else if (ReadChar(f, c, fnt)) break;
        return 0;
      case '\n':
      case '\0':
      case ' ':
      case '\t':
        break;
      default:
        printf("Непонятная строчка %d.\n", n);
        return 0;
    } /* switch (s[0]) */
  } /* for (;;) */
}


void Dump(FILE *f)
{ int i=0, j;
  char *s;
  puts("Начинаю записывать chargen.c");
  while (s=tblPreface[i++])
    fputs(s, f);

  for (i=0; i<3; i++)
  { fputs(" { ", f);
    for (j=0; j<256; j++)
    { fprintf(f, "%d", tblFontSize[i][j]);
      if ((j&15)==15)
      { if (j==255) fputs("  /* f0..ff */\n  }", f);
        else fprintf(f, ", /* %02x..%02x */\n    ", j&0xf0, j);
      } else fputs(", ", f);
    }
    if (i!=2) fputs(",\n ", f);
  }
  fputs("\n};\n\n\n", f);

  for (i=0; i<3; i++)
    for (j=0; j<256; j++)
      if (tblFontSize[i][j])
      { fprintf(f, "static TCharacter c%02xn%d =\n{ ", j, i);
        DumpChar(f, j, i);
        fputs("\n};\n\n", f);
      }

  fputs("\nTCharacter *tblFont [3][256] =\n{", f);

  for (i=0; i<3; i++)
  { fputs(" { ", f);
    for (j=0; j<256; j++)
    { if (tblFontSize[i][j]) fprintf(f, "&c%02xn%d", j, i);
      else fputs("0", f);
      if ((j&7)==7)
      { if (j==255) fputs("  /* f8..ff */\n  }", f);
        else fprintf(f, ", /* %02x..%02x */\n    ", j&0xf8, j);
      } else fputs(", ", f);
    }
    if (i!=2) fputs(",\n ", f);
  }
  fputs("\n};\n\n\n", f);

  puts("Создан файл chargen.c");
}


int main(int argc, char *argv[])
{ FILE *fi = fopen("chargen.txt", "r");
  if (fi!=NULL)
  { FILE *fo = fopen("chargen.c", "w");
    if (fo!=NULL)
    { InitTables();
      if ((argc==2) && (strcmp(argv[1], "-v")==0))
        flgVerbose = 1;
      else
        flgVerbose = 0;
      if (Read(fi)) Dump(fo);
      fclose(fo);
    } else puts("Ошибка создания файла chargen.c");
    fclose(fi);
  } else puts("Ошибка открытия файла chargen.txt");
  return 0;
}

