
/* getcharimg.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 12 апреля 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Создать CHARGEN.TXT на основе знакогенератора МК-161,
 * опубликованного НПП "СЕМИКО" под лицензией BSD.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define NUM_BYTES (32500)


FILE *fi, *fo;
unsigned char s[100], *p;
unsigned char sLabel[30];

int nLine;

unsigned char bSize[256]; /* ширина символов */
unsigned char iHeight[3] = { 1, 2, 3 }; /* высота символа в байтах */
unsigned char iLines[3] = { 8, 14, 19 }; /* высота символа в точках */

int iPicNum[256]; /* номера картинок, для каждого символов */

unsigned char *pName[256]; /* метки картинок */
int pGlyph[256], iBC; /* адреса картинок */

unsigned char bBytes[NUM_BYTES]; /* массив для картинок */


void Init(void)
{ int i;
  iBC = 0;
  for (i=0; i<256; i++)
    bSize[i] = 0,
    pName[i] = NULL,
    pGlyph[i] = -1,
    iPicNum[i] = -1;
  for (i=0; i<NUM_BYTES; i++)
    bBytes[i] = 0;
}


void DeInit(void)
{ int i;
  iBC = 0;
  for (i=0; i<256; i++)
  { if (pName[i] !=NULL) free(pName[i]),  pName[i]  = NULL;
    bSize[i] = 0;
    pGlyph[i] = -1;
    iPicNum[i] = -1;
  }
}


int IsSpace(unsigned char c)
{ return (c==' ') || (c=='\t') || (c=='\n') || (c=='\r');
}


int IsHexCh(unsigned char c)
{ if ((c>='0') && (c<='9')) return 1;
  if ((c>='a') && (c<='f')) return 1;
  if ((c>='A') && (c<='F')) return 1;
  if ((c=='h') || (c=='H')) return 1;
  return 0;
}


void Space(void)
{ while (IsSpace(*p)) p++;
}


int Literal(const char *z)
{ unsigned char *q = p;
  while (*z)
    if (*q++ != *z++) return 0;
  p = q;
  return 1;
}


int ParseNum(void)
{ unsigned char z[20], ch;
  unsigned char *q = z;
  int flgHex, n = 0;
  Space();
  if (Literal("PRO")) return 31;
  if (IsHexCh(*p)==0)
  { printf(" *** Здесь требуется число:\n%s\n", p);
    return -1;
  }
  while (IsHexCh(*p))
    *q++ = *p++;
  *q = '\0';
  ch = *(q-1);
  if ((ch=='h') || (ch=='H')) flgHex = 1, *(q-1) = '\0';
  else flgHex = 0;
  q = z;
  while (*q)
  { ch = *q++;
    if (flgHex) n*=16; else n*=10;
    if ((ch>='0') && (ch<='9')) n += ch-'0';
    else if (flgHex && (ch>='a') && (ch<='f')) n += ch - 'a' + 10;
    else if (flgHex && (ch>='A') && (ch<='F')) n += ch - 'A' + 10;
    else
    { printf(" *** Нераспознанное число '%s'.\n", z);
      return -1;
    }
  }
  return n;
}


int ParseChar(void)
{ Space();
  if (*p=='\'')
  { int n = p[1];
    if (p[2]=='\'')
    { p += 3;
      return n;
    }
    printf(" *** Нераспознанный символ:\n%s\n", p);
    return -1;
  } else return ParseNum();
}


int ParseByte(void)
{ Space();
  if (*p=='\0')
  { while (*p=='\0')
    { fgets(s, 100, fi);
      nLine++; /* Нет проверки на конец файла. */
      p = s;
      Space();
    }
    if (Literal(".DB")) Space();
    else
    { printf(" *** Строка описания должна начинаться с .DB:\n%s\n", p);
      return -1;
    }
  } else if (Literal(",")) Space();
  else
  { printf(" *** Между байтами должна быть запятая:\n%s\n", p);
    return -1;
  }
  return ParseNum();
}


void GetLabel(void)
{ unsigned char *q = sLabel;
  while ((*p) && (*p!='-') && (*p!=':') && (*p!=' '))
    *q++ = *p++; /* Нет проверки на переполнение. */
  *q++ = '\0';
}


int SearchLabel(void)
/* Найти метку в массиве pName[] */
{ int i=0;
  GetLabel();
  while (pName[i]!=NULL)
    if (strcmp(pName[i], sLabel)==0) return i;
    else i++; /* Нет проверки на переполнение. */
  return -1;
}


void ParseLabelDef(int ch)
{ int n = SearchLabel();
  if (n<0)
  { n = 0; /* Регистрируем новую метку */
    while (pName[n]!=NULL) n++; /* Нет проверки переполнения. */
    pName[n] = malloc(strlen(sLabel)+1);
    strcpy(pName[n], sLabel); /* Нет проверки нехватки памяти. */
  }
  if (iPicNum[ch]!=-1)
  { printf(" *** Двойное определение символа %d.\n", ch);
    return;
  }
  iPicNum[ch] = n;
}


void ReadFont(int iFont)
{ int flgGetLine = 1;
  static char sFont[] = "FONT_0:", sKS[] = "KS0_";
  sFont[5] = '0'+iFont;
  sKS[2] = '0'+iFont;
  nLine = 0;
  while(!feof(fi))
  { if (flgGetLine)
    { fgets(s, 100, fi);
      p = s;
      nLine++;
      flgGetLine = 0;
    }
    Space();
    if ((*p=='\0') || (*p==';')) ;
    else if (Literal(sFont))
    { fgets(p = s, 100, fi);
      nLine++; /* Мы и так знаем высоту буквы, игнорировать строчку. */
    } else if (Literal(".DB"))
    { int ch;
      Space();
      if (Literal("KONT") || Literal("0FFH")) break; /* Конец первой части */
      else
      { ch = ParseChar();
        if (ch<0) return;
        if (Literal(",")) bSize[ch] = ParseNum();
        else
        { printf(" *** Ошибка в определении символа %d в строке %d:\n%s\n",
            ch, nLine, p);
          return;
        }
       fgets(p = s, 100, fi);
       nLine++;
       Space();
       if (Literal(".DW"))
       { Space();
         if (Literal(sKS)) ParseLabelDef(ch);
         else
         { printf(" *** Ожидалась метка %s в строке %d:\n%s\n",
             sKS, nLine, p);
           return;
         }
       } else
       { printf(" *** Ожидалась команда .DW в строке %d:\n%s\n",
           nLine, p);
         return;
       }
      } /* if KONT/0FFH */
    } else
    { printf(" *** Неизвестная команда в строке %d:\n%s\n", nLine, p);
      return;
    }
    flgGetLine = 1;
  } /* while !eof */
  iBC = 0;
  while(!feof(fi))
  { fgets(s, 100, fi);
    p = s;
    nLine++;
    Space();
    if ((*p=='\0') || (*p==';')) ;
    else if (Literal(sKS))
    { int n = SearchLabel(), sz;
      if (n<0)
      { printf(" *** Неизвестный глиф %s в строке %d:\n%s\n",
          sLabel, nLine, p);
        return;
      }
      if (*p!=':')
      { printf(" *** Неправильная метка %s в строке %d:\n%s\n",
          sLabel, nLine, p);
        return;
      }
      pGlyph[n] = iBC;
    } else if (Literal(".DB"))
    { Space();
      while (1)
      { int q = ParseNum();
        if (q<0) return;
        bBytes[iBC++] = q; /* Нет проверки переполнения */
        Space();
        if (*p==0) break;
        if (*p!=',')
        { printf(" *** Ожидалась запятая в строке %d:\n%s\n",
            nLine, p);
          return;
        }
        p++;
        Space();
        if (*p==0)
        { printf(" *** Строка %d заканчивается запятой.\n", nLine);
          return;
        }
      } /* while(1) */
    } /* if */
  } /* while !eof */
}


void WriteFont(int iFont)
{ int ch, h = iHeight[iFont];
  for (ch=0; ch<256; ch++)
  { int n = iPicNum[ch];
    int w = bSize[ch];
    if (n>=0)
    { int g = pGlyph[n];
      int ln = iLines[iFont];
      int m, i, j;
      if (w==0)
      { printf(" *** Символ $02x нулевой ширины.\n", ch);
        return;
      }
      if (pName[n]==NULL)
      { printf(" *** Безымянный символ %02x.\n", ch);
        return;
      }
      if (g==-1)
      { printf(" *** Нет картинки символа %s с кодом %02x.\n",
          pName[n], ch);
        return;
      }
      fprintf(fo, "\n:%02x,%d\n", ch, iFont);
      j = 0;
      m = 1;
      while (ln>0)
      { for (i=0; i<w; i++)
          if (m & bBytes[g+j+i*h]) putc('x', fo);
          else putc('.', fo);
        putc('\n', fo);
        ln--;
        m <<= 1;
        if (m>0x80) m = 1, j++;
      } /* while ln */
    } else if (w!=0)
    { printf(" *** Не найден глиф символа %02x.\n", ch);
      return;
    }
  } /* for i */
  fputs("\n---\n", fo); /* По хорошему, нужно лишь в конце шрифта 2. */
}


void Convert(const char *sFileIn, const char *sFileOut, int iFont)
{ fi = fopen(sFileIn, "r");
  if (fi==NULL)
  { printf(" *** Не могу открыть файл '%s'.\n", sFileIn);
    return;
  }
  printf("Считываю шрифт из файла '%s'.\n", sFileIn);
  Init();
  ReadFont(iFont);
  fclose(fi);
  fo = fopen(sFileOut, "w");
  if (fo!=NULL)
  { printf("Сохраняю шрифт в файл '%s'.\n", sFileOut);
    WriteFont(iFont);
    fclose(fo);
  } else printf(" *** Не могу создать файл '%s'.\n", sFileOut);
  DeInit();
}


int main(void)
{ puts("Преобразование шрифтов МК-161 в шрифты ВК-6.");
  Convert("f0_1_01.asm", "chargen0.txt", 0);
  Convert("f1_1_01-mod.asm", "chargen1.txt", 1);
  Convert("f2_1_00.asm", "chargen2.txt", 2);
  puts("Сделано!");
  return 0;
}

