
/* reg.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 5 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Регистры и основные регистровые команды.
 * Регистры функций.
 *
 * Функциональный узел ЭКВМ: регистры данных и функций.
 * Функциональный узел ЭКВМ: блок питания.
 *
 * В отличии от ЭКВМ, операция Мr может вызвать переполнение,
 * если значение RX не укладывается в 12-разрядную мантиссу.
 */

#include <math.h>
#include <stdlib.h>

#include "cpu.h"
#include "reg.h"
#include "display.h"
#include "decomp.h"
#include "comment.h"
#include "buffer.h"
#include "clock.h"
#include "supp.h"
#include "keyboard.h"
#include "os.h"


real ldR[1000];
unsigned char bR[7168];

int iNetNum=0; /* R9049: задание сетевого номера    */
int iStyle=0;  /* R9110: установка разрядности      */
int iBinary=0; /* R9100: разрядность двоичных чисел */
int iPC2=0;    /* R9042: индексный регистр          */

static long lRND0 = 0; /* R9046: инициализация К СЧ */

void FnInit(void)
/* Инициализация всех регистров функций (драйверов). */
{ iPC2 = 0;
  iNetNum = 0;
  iStyle = 0;
  iBinary = 0;
  iBufInd0 = iBufInd1 = 0;
  OsInit();
  DrawInit();
  ClockInit();
}


static void R9040w(void)
/* Вычисление чётности байта. */
{ int c=0, i;
  for (i=MkByte(RX); i; i>>=1)
    if (i&1) c++;
  Push(c);
}


static void R9044r(void)
/* Чтение содержимого памяти программ. */
{ Push(bProg[iPC2++]);
  if (iPC2>9999) iPC2=0;
}


static void BinTrim(real *pd, int *pi)
/* Ограничить число в RX и выдать число байт. */
{ real m;
  switch (iBinary)
  { case 4:  m = 4294967295.L;  *pi = 4;  break;
    case 3:  m = 16777215.L;    *pi = 3;  break;
    case 2:  m = 65535.L;       *pi = 2;  break;
    default: m = 255.L;         *pi = 1;  break;
  }
  if (RX<=0) *pd = 0; else if (RX>=m) *pd = m; else *pd = floor(RX);
}


static void HexOut(real d, int depth)
{ int n = (int) fmod(d, 256);
  if (depth)
  { HexOut((d-n)/256, depth-1);
    CommentStoreCh(' ');
  }
  CommentStoreCh(strHexUp[n/16]);
  CommentStoreCh(strHexUp[n%16]);
}


static void HexWrite(real d, int depth)
{ int n = (int) fmod(d, 256);
  if (depth) HexWrite((d-n)/256, depth-1);
  BufWrite(n);
}


static void R9102w(void)
/* Вывод двоичного числа в строку комментариев. */
{ int i;
  real d;
  BinTrim(&d, &i);
  CommentStoreCh(13);
  HexOut(d, i-1);
}


static void R9103w(void)
/* Преобразование числа в последовательность байтов в двоичном виде. */
{ int i;
  real d;
  BinTrim(&d, &i);
  BufStart();
  HexWrite(d, i-1);
}


static void R9103r(void)
/* Преобразование числа из последовательности байтов в двоичном виде. */
{ int i = 0;
  real d = 0;
  BufStart();
  do d = d*256 + BufRead();
  while (++i < iBinary);
  Push(d);
}


static void R9046w(void)
/* Инициализация генератора случайных чисел. */
{ lRND0 = Trunc(65535);
  srand(lRND0);
}


void FnM(int fn)
/* Записать RX в регистр функций с номером fn. */
{ switch (fn)
  { case 9000: R9000w();            break; /* Запись нач. координат объекта */
    case 9001: iGrAttr=Trunc(7);    break; /* Запись атрибутов вывода       */
    case 9002: iGrDist=Trunc(15);   break; /* Запись расстояния между симв. */
    case 9003: iGrFont=Trunc(2);    break; /* Запись номера шрифта          */
    case 9004: iGrInd=Trunc(1023);  break; /* Запись индексного регистра    */
    case 9005: bScreen[iGrInd]=MkByte(RX);  break;
    case 9006: R9006w();            break; /* Вывод в экранный буфер, инк.  */
    case 9007: Scroll(0);           break; /* Прокрутка экрана              */
    case 9008: Scroll(255);         break; /* Прокрутка экрана              */
    case 9010: R9010w();            break; /* Инициализация граф. экрана    */
    case 9011: R9011w();            break; /* Вывод точки                   */
    case 9012: R9012w();            break; /* Вывод линии                   */
    case 9013: R9013w();            break; /* Вывод прямоугольника          */
    case 9014: R9014w();            break; /* Вывод рамки                   */
    case 9015: R9015w();            break; /* Вывод граф. образа, из прог.  */
    case 9016: R9016w();            break; /* Вывод граф. образа, данные    */
    case 9020: R9020w();            break; /* Вывод символа                 */
    case 9021: R9021w();            break; /* Вывод строки символов, прог.  */
    case 9022: R9022w();            break; /* Вывод строки символов, данные */
    case 9023: R9023w();            break; /* Вывод числа с ест. или пл. з. */
    case 9024: R9024w();            break; /* Вывод числа в форматир. виде  */
    case 9025: R9025w();            break; /* Строка комментариев: символ   */
    case 9026: R9026w();            break; /* Строка комментариев: из прог. */
    case 9027: R9027w();            break; /* Строка комментариев: данные   */
    case 9029: KbdReset();          break; /* Клавиатура: очистка буфера    */
    case 9030: iBufInd0=MkByte(RX); break; /* Буфер: индексный регистр 0    */
    case 9031: iBufInd1=MkByte(RX); break; /* Буфер: индексный регистр 1    */
    case 9032: R9032w();            break; /* Буфер: запись по инд. 0       */
    case 9033: R9033w();            break; /* Буфер: запись по инд. 1       */
    case 9034: R9034w();            break; /* Буфер: запись по инд. 1, инк. */
    case 9035: R9035w();            break; /* Буфер: преобр. числа с е/п з. */
    case 9036: R9036w();            break; /* Буфер: преобр. числа в форм.  */
    case 9037: R9037w();            break; /* Буфер: во внутренний формат   */
    case 9038: R9038w();            break; /* Буфер: преобразование в FLOAT */
    case 9039: R9039w();            break; /* Буфер: код ком. в мнемонику   */
    case 9040: R9040w();            break; /* Вычисление чётности байта     */
    case 9041: Push(iPC);           break; /* Чтение счётчика адреса        */
    case 9042: iPC2=MkByte(RX);     break; /* Запись индексного регистра    */
    case 9045: flgGrad=Trunc(2);    break; /* Размерность аргумента триг.ф. */
    case 9046: R9046w();            break; /* Инициализация генератора СЧ   */
    case 9049: iNetNum=MkByte(RX);  break; /* Задание сетевого номера       */
    case 9050: R9050w();            break; /* Регистр таймера 0             */
    case 9051: R9051w();            break; /* Регистр таймера 1             */
    case 9052: R9052w();            break; /* Формирование звук. сигнала    */
    case 9053: R9053w();            break; /* -"- при нажатии на кнопку     */
    case 9056: iDate=Trun1(31);     break; /* Часы: корректировка числа     */
    case 9057: iMonth=Trun1(12);    break; /* Часы: корректировка месяца    */
    case 9058: iYear=Trunc(9999);   break; /* Часы: корректировка года      */
    case 9059: iDay=Trun1(7);       break; /* Часы: корректировка дня нед.  */
    case 9100: iBinary=Trunc(4);    break; /* Разрядность двоичных чисел    */
    case 9102: R9102w();            break; /* Вывод двоичного числа в ком.  */
    case 9103: R9103w();            break; /* Преобразование числа, в буфер */
    case 9110: iStyle=Trunc(1);     break; /* Установка разрядности         */
  }
}


void FnRM(int fn)
/* Считать в стек регистр функций с номером fn. */
{ switch (fn)
  { case 9000: R9000r();          break; /* Чтение нач. координат объекта */
    case 9001: Push(iGrAttr);     break; /* Чтение атрибутов вывода       */
    case 9002: Push(iGrDist);     break; /* Чтение расстояния между симв. */
    case 9003: Push(iGrFont);     break; /* Чтение номера шрифта          */
    case 9004: Push(iGrInd);      break; /* Чтение индексного регистра    */
    case 9005: Push(bScreen[iGrInd]);  break;
    case 9006: R9006r();          break; /* Чтение экранного буфера, инк. */
    case 9011: R9011r();          break; /* Состояние точки на гр. экране */
    case 9028: R9028r();          break; /* Чтение состояния клавиатуры   */
    case 9029: Push(KbdGet());    break; /* Чтение кода нажатой кнопки    */
    case 9030: Push(iBufInd0);    break; /* Буфер: индексный регистр 0    */
    case 9031: Push(iBufInd1);    break; /* Буфер: индексный регистр 1    */
    case 9032: R9032r();          break; /* Буфер: чтение по инд. 0       */
    case 9033: R9033r();          break; /* Буфер: чтение по инд. 1       */
    case 9034: R9034r();          break; /* Буфер: чтение по инд. 1, инк. */
    case 9037: R9037r();          break; /* Буфер: из внутреннего формата */
    case 9038: R9038r();          break; /* Буфер: пр-е из формата FLOAT  */
    case 9041: Push(iPC);         break; /* Значение счётчика адреса      */
    case 9042: Push(iPC2);        break; /* Чтение индексного регистра    */
    case 9043: Push(bProg[iPC2]); break; /* Чтение содержимого пам. прог. */
    case 9044: R9044r();          break; /* Чтение содержимого пам. прог. */
    case 9045: Push(flgGrad);     break; /* Размерность аргумента триг.ф. */
    case 9046: Push(lRND0);       break; /* Инициализация генератора К СЧ */
    case 9049: Push(iNetNum);     break; /* Чтение сетевого номера        */
    case 9050: Push(lTimer0);     break; /* Регистр таймера 0             */
    case 9051: Push(lTimer1);     break; /* Регистр таймера 1             */
    case 9052: R9052r();          break; /* Формирование звук. сигнала    */
    case 9053: R9053r();          break; /* -"- при нажатии на кнопку     */
    case 9055: R9055r();          break; /* Часы: чтение текущего времени */
    case 9056: Push(iDate);       break; /* Часы: чтение числа            */
    case 9057: Push(iMonth);      break; /* Часы: чтение месяца           */
    case 9058: Push(iYear);       break; /* Часы: чтение года             */
    case 9059: Push(iDay);        break; /* Часы: чтение дня недели       */
    case 9100: Push(iBinary);     break; /* Разрядность двоичных чисел    */
    case 9103: R9103r();          break; /* Считать двоичное число, буфер */
    case 9110: Push(iStyle);      break; /* Чтение разрядности            */
    case 9115: Push(3.6);         break; /* Напряжение источника питания  */
    case 9116: Push(6);           break; /* Номер модели ЭКВМ             */
    case 9117: Push(NUM_VERSION); break; /* Номер версии ПО               */
  }
}


void Modify(int r, int flgShift)
/* Модифицировать регистр перед обращением к StoreK / RecallK.
   Если flgShift отличен от нуля, произвести автодекремент/автоинкремент.
 */
{ real x = ldR[r];
  if (flgShift)
    if (r<=3) --x; else if (r<=6) ++x;
  if (x<=0) ldR[r] = 0;
  else ldR[r] = floor(x);
}


static void StoreR(int r)
/* Записать RX в регистр с двоичным порядковым номером r. */
{ if (r<1000) ldR[r] = RX;
  else if (r<8168) bR[r-1000] = MkByte(RX);
  else if ((r>=9000) && (r<10000)) FnM(r);
}


static void RecallR(int r)
/* Считать в стек регистр с двоичным порядковым номером r. */
{ if (r<1000) Push(ldR[r]);
  else if (r<8168) Push(bR[r-1000]);
  else if ((r>=9000) && (r<10000)) FnRM(r);
}


static void StoreK(int r) /* Записать RX в рег., номер которого в рег. r. */
  { if (ldR[r]<10000) StoreR((int)ldR[r]); }

static void RecallK(int r) /* Считать в стек рег., номер которого в рег. r. */
  { if (ldR[r]<10000) RecallR((int)ldR[r]); }


void M(int r)
  { if (Guard12()) State(MnemoM(r)),  ldR[r] = RX; }

void RM(int r)
  { if (Guard()) State(MnemoRM(r)),  Push(ldR[r]); }

void PM(int bcd)
  { if (Guard12()) State(MnemoPM(bcd)),  ldR[Bin(bcd)] = RX; }

void PRM(int bcd)
  { if (Guard()) State(MnemoPRM(bcd)),  Push(ldR[Bin(bcd)]); }

void KM(int r)
  { if (Guard12()) State(MnemoKM(r)),  Modify(r, 1),  StoreK(r); }

void KRM(int r)
  { if (Guard()) State(MnemoKRM(r)),  Modify(r, 1), RecallK(r); }

void PPM(int bcd1, int bcd2)
  { if (Guard12()) State(MnemoPPM(bcd1, bcd2)),  StoreR(Bin2(bcd1, bcd2)); }

void PPRM(int bcd1, int bcd2)
  { if (Guard()) State(MnemoPPRM(bcd1, bcd2)),  RecallR(Bin2(bcd1, bcd2)); }


void PKM(int bcd)
{ if (Guard12())
  { int r = Bin(bcd);
    State(MnemoPKM(bcd));
    Modify(r, 0);
    StoreK(r);
  }
}


void PKRM(int bcd)
{ if (Guard())
  { int r = Bin(bcd);
    State(MnemoPKRM(bcd));
    Modify(r, 0);
    RecallK(r);
  }
}


