

/* supp.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 10 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Вспомогательные функции.
 */

#include <math.h>

#include "cpu.h"
#include "supp.h"


char strHex[]   = "0123456789abcdef",
     strHexUp[] = "0123456789ABCDEF";

static unsigned char strNumber[15];


unsigned char Up(unsigned char c)
/* Преобразовать букву в верхний регистр, учитывая русскую кодировку ВК-6. */
{ if ((c>='a') && (c<='z')) return c-'a'+'A';
  else if ((c>=0xa0) && (c<=0xaf)) return c-0xa0+0x80; /* а..п */
  else if ((c>=0xe0) && (c<=0xef)) return c-0xe0+0x90; /* р..я */
  else if (c==0xf1) return 0xf0; /* ё */
  return c;
}


int IsSpace(unsigned char c)
  { return (c==' ') || (c=='\t') || (c=='\n') || (c=='\r'); }

int IsDigit(unsigned char c)
  { return (c>='0') && (c<='9'); }

void Space(const unsigned char **s) /* Пропустить пробелы во входной строке. */
  { if (*s) while (IsSpace(**s)) (*s)++; }


int Bin(int bcd)
  { return (bcd/16)*10 + bcd%16; }

int Bin2(int bcd1, int bcd2)
  { return (bcd1/16)*1000 + (bcd1%16)*100 + (bcd2/16)*10 + bcd2%16; }

void Push(real x)
  { RT = RZ;  RZ = RY;  RY = RX;  RX = x; }


real Abs(real x)
{ if (x<0) return -x;
  return x;
}


int AbsI(int x)
{ if (x<0) return -x;
  return x;
}


void SwapI(int *x, int *y)
{ int z = *x;
  *x = *y;
  *y = z;
}


int MkByte(real ld)
{ if (ld<=0) return 0;
  if (ld>=255) return 255;
  return (int) floor(ld);
}


long Trunc(long x)
{ if (RX<=0) return 0;
  if (RX>=x) return x;
  return (long) floor(RX);
}


long Trun1(long x)
{ if (RX<=1) return 1;
  if (RX>=x) return x;
  return (long) floor(RX);
}


long TruncY(long x)
{ if (RY<=0) return 0;
  if (RY>=x) return x;
  return (long) floor(RY);
}


long Trun1Y(long x)
{ if (RY<=1) return 1;
  if (RY>=x) return x;
  return (long) floor(RY);
}


long TruncZ(long x)
{ if (RZ<=0) return 0;
  if (RZ>=x) return x;
  return (long) floor(RZ);
}


unsigned char *ToNumMK(real x)
/* Преобразовать x в восьмибайтовое представление числа в ЭКВМ. */
{ static unsigned char buf[8];
  int i = 1;
  if (x<0) buf[0] = 0x80, x = -x; else buf[0] = 0;
  if (x!=0)
  { real r = log10(x);
    int iPoryadok = floor(r);
    if (iPoryadok>=0) buf[7] = iPoryadok; else buf[7] = 256+iPoryadok;
    r = x / pow(10, iPoryadok+1);
    while (i<7)
    { int d = floor(r*100);
      r = r*100 - d;
      if (Abs(r)<5E-11L) r = 0; /* двоичная плавучка не так надёжна */
      else if ((d<99) && (Abs(r-1)<=5E-11L)) d++, r = 0;
      if ((i==6) && (d<99) && (r>=0.5)) d++;  /* худо-бедно округляем */
      buf[i++] = (d/10)*16 + (d%10);
    }
  } else while (i<8) buf[i++] = 0;
  return buf;
}


real FromNumMK(const unsigned char *s)
/* Считать восьмибайтовое представление дестичного числа в ЭКВМ. */
{ real r = 0;
  int i = 1, iPoryadok = s[7];
  if (iPoryadok>127) iPoryadok -= 256;
  if ((iPoryadok>99) && (iPoryadok<-99)) return 0;
  while (i<7)
  { int d1 = s[i]/16;
    int d2 = s[i++]%16;
    if ((d1>9) || (d2>9)) return 0;
    r = r*100 + d1*10+d2;
  }
  r *= pow(10, iPoryadok-11);
  if (s[0]==0) return r;
  if (s[0]==0x80) return -r;
  return 0;
}


unsigned char *ToFloat(real x)
{ static unsigned char s[4];
  real r;
  int iPoryadok, m;
  if (x<0) x = -x, s[3] = 0x80; else s[3] = 0;
  r = 256*frexp(x, &iPoryadok);
  iPoryadok += 127-1;
  m = (int) floor(r);
  r -= m;
  if (Abs(r)<1E-10L) r = 0;
  else if ((m<255) && (Abs(r-1)<=1E-10)) m++, r = 0;
  s[2] = m & 0x7f;
  if (iPoryadok&1) s[2] |= 0x80;
  s[3] |= (iPoryadok>>1);
  r *= 256;
  m = (int) floor(r);
  r -= m;
  if (Abs(r)<1E-10L) r = 0;
  else if ((m<255) && (Abs(r-1)<=1E-10)) m++, r = 0;
  s[1] = m;
  m = (int) floor(256*r);
  if ((m<255) && (r-m>=0.5)) m++; /* Худо-бедно округляем */
  s[0] = m;
  return s;
}


real FromFloat(const unsigned char *s)
{ real r =  (s[2] | 0x80)*65536L + s[1]*256L + s[0];
  int i = 2, iPoryadok = (s[3] & 0x7f)<<1;
  if (s[2]&0x80) iPoryadok++;
  r *= pow(2, iPoryadok-127-23);
  if (s[3]&0x80) return -r;
  return r;
}


static StoreFrac(unsigned char **s, real r, int n)
/* Записать в строку *s дробную часть r длиной n. */
{ int i, d;
  real q = 5*pow(10, -n-1);
  if (r>=q)
  { *(*s)++ = ',';
    for (i=0; i<n; i++)
      if (r>=q)
      { r *= 10;  q *= 10;
        if (i+1==n)
        { d = floor(r+0.5); /* округление для бедных */
          if (d>9) d=9;
        } else d = floor(r), r -= d;
        *(*s)++ = strHex[d];
      } else break;
  }
}


static void StoreInteger(unsigned char **s, real r, int *pn)
/* Записать в строку *s целую часть r, уменьшая кол-во цифр в *pn.  */
{ real q = floor(r/10);
  if (q>0) StoreInteger(s, q, pn);
  *(*s)++ = strHex[(int) floor(r-q*10)];
  (*pn)--;
}


static unsigned char *StoreFloat(unsigned char *s, real x)
/* Записать в строку s положительное число x с плавающей запятой. */
{ real r = log10(x);
  int iPoryadok = floor(r), d;
  r = x / pow(10, iPoryadok);
  d = floor(r);  r -= d;
  *s++ = strHex[d];
  StoreFrac(&s, r, 7);
  if (iPoryadok)
  { *s++ = 'E';
    if (iPoryadok<0) *s++ = '-', iPoryadok = -iPoryadok;
    if (iPoryadok>99) *s++ = strHex[iPoryadok/100], iPoryadok %= 100;
    *s++ = strHex[iPoryadok/10];
    *s++ = strHex[iPoryadok%10];
  }
  return s;
}


unsigned char *NumToStr(real x)
/* Преобразовать в строку число с естественной или плавающей запятой. */
{ unsigned char *s = strNumber;
  if (x<0) *s++ = '-', x = -x;
  if (x<1E-99) x = 0;
  if ((x>=1) && (x<1E8))
  { int n = 8;
    real r = floor(x);
    x -= r;
    StoreInteger(&s, r, &n);
    StoreFrac(&s, x, n);
  } else s = StoreFloat(s, x);
  *s = '\0';
  return strNumber;
}


unsigned char *NumToStrFix(real x, int c, int d)
/* Преобразовать в строку число в форматированном виде, если это возможно. */
{ unsigned char *s = strNumber, *p = strNumber+c;
  real r = Abs(x);
  real t = floor(r); /* целая часть */
  if (c>8) c = 8;
  if (c+d>8) d = 8-c;
  if (r<5*pow(10, -d-1)) x = r = t = 0;
  if (r>=pow(10, c)) return NumToStr(x);
  r -= t;           /* дробная часть */
  if ((d==0) && (r>=0.5)) t++; /* округление */
  do
  { int dgt = (int) fmod(t, 10);
    t = (t-dgt)/10;
    *p-- = strHex[dgt];
  } while ((p>strNumber) && (t>=1));
  if (x<0) *p-- = '-';
  while (p>=strNumber)
    *p-- = ' ';
  p = strNumber+c+1;
  if (d!=0)
  { StoreFrac(&p, r, d);
    while (p<strNumber+c+d+2)
      *p++ = ' ';
  } else *p++ = ' ';
  *p = '\0';
  return strNumber;
}


