
/* vk6.c
 * ВК-6 v0.30
 * (c) Васильев И.В., Москва, 5 марта 2010 г.
 *
 * Программа распространяется под лицензией GNU GPL v3.
 *
 * Интерфейс ВК-6 с пользователем и операционной системой.
 *
 * Этот модуль также переводит русские буквы в кодировку ВК-6 и обратно.
 * Преобразуйте исходный текст модуля в кодировку, принятую в вашей
 * операционной системе.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vk6.h"
#include "cpu.h"
#include "exec.h"
#include "display.h"
#include "decomp.h"
#include "comment.h"
#include "supp.h"
#include "reg.h"
#include "os.h"


int iExitCode = 0;
/*  00 Успех
    01 Ошибка командной строки
 */

int optDialog = 0,  /* -i */
    optClear = 0,   /* -c */
    optRot = 0,     /* -r */
    optNegativ = 0, /* -n */
    optX1 = 0;      /* -x */

static int flgLoadedMKP = 0;
int flgLoadedMKI = 0;


static const char *alRusUp[] =
{ "А", "Б", "В", "Г", "Д", "Е", "Ж", "З",
  "И", "Й", "К", "Л", "М", "Н", "О", "П",
  "Р", "С", "Т", "У", "Ф", "Х", "Ц", "Ч",
  "Ш", "Щ", "Ъ", "Ы", "Ь", "Э", "Ю", "Я", 0
}, *alRusL1[] =
{ "а", "б", "в", "г", "д", "е", "ж", "з",
  "и", "й", "к", "л", "м", "н", "о", "п", 0
}, *alRusL2[] =
{ "р", "с", "т", "у", "ф", "х", "ц", "ч",
  "ш", "щ", "ъ", "ы", "ь", "э", "ю", "я", 0
}, *alRusYo[] =
{ "Ё", "ё", 0
}, *alSpec[] =
{ "÷", "≥", "≠", "×", "↵", "↻", 0,   "↔",
  0,   "π", 0,   0,   0,   "²", "x", "y",
  "↵", "↤", "⇤", "⇥", "←", "→", "↑", "↓",
  "↖", "↘", "↔", 0,   0,   0,   0,   0
};

static char *alBlocks[2][4] = { { "█", "▄", "▀", " " },
                                { " ", "▀", "▄", "█" } };


static void Help(void)
{ puts("Опции калькулятора ВК-6:\n"
       " -i  перевести ВК-6 в режим диалога с пользователем\n"
       " -c  очищать экран перед каждым выводом\n"
       " -r  вращать граф. экран (требует более 128 cpl)\n"
       " -n  отображать граф. экран в негативе\n"
       " -x  при выводе стека распечатывать регистр X1\n"
       " -h  вывести эту подсказку\n");
  exit(iExitCode=0);
}


static void AbortCmdLine(void)
{ puts("Примеры использования:\n"
       "   vk6 -ic       полноэкранный диалог с ВК-6\n"
       "   vk6 pp.mkp    запуск программы pp.mkp на ВК-6\n");
  exit(iExitCode=1);
}


static void Options(const unsigned char *s)
{ unsigned char c=*s++;
  if (c!='-') AbortCmdLine();
  while (c=*s++)
    switch (c)
    { case 'i': optDialog = 1;   break;
      case 'c': optClear = 1;    break;
      case 'r': optRot = 1;      break;
      case 'n': optNegativ = 1;  break;
      case 'x': optX1 = 1;       break;
      case 'h':
      case '?': Help();
      default:  AbortCmdLine();
    }
}


static void LoadFailed(const char *s, int i)
{ printf("Неверный формат или ошибка чтения, файл загружен частично.\n"
         "Число %s: %d\n", s, i);
}


void LoadMKP(FILE *f, long sz)
{ int c = fgetc(f);
  flgLoadedMKP = 0;
  if ((sz>=c*200+1) && (sz<=20001))
  { int i = Bin2(c, 0);
    int i0 = i;
    printf("Начальный адрес: %04d\n", i0);
    while (i<10000)
    { int d=fgetc(f);
      if (d==EOF) break;
      if ((c=fgetc(f))==EOF) break;
      if ((c!=0) && (c!=1)) { LoadFailed("шагов", i);  return; }
      bProg[i++]=d;
    }
    i -= i0;
    printf("Загружен");
    switch (i%10)
    { case 1:  printf(" %d шаг.\n", i);
               break;
      case 2:
      case 3:
      case 4:  printf("о %d шага.\n", i);
               break;
      default: printf("о %d шагов.\n", i);
    }
    iPC = i0;
    flgLoadedMKP=1;
  } else OsAlert("vk6", "Неверный формат программного файла.");
}


static void LoadMKL(FILE *f, long sz)
{ OsAlert("vk6", "Загрузка файлов MKL не предусмотрена.");
}


static void LoadMKD(FILE *f, long sz)
{ if (sz==8000)
  { unsigned char s[8];
    int i = 0;
    while (i<1000)
    { size_t n = fread(s, 1, 8, f);
      if ((n!=8) || ferror(f))
      { LoadFailed("десятичных регистров", i);  
        return;
      }
      ldR[i++] = FromNumMK(s);
    }
    puts("Загружена тысяча десятичных регистров.");
  } else OsAlert("vk6", "Неверный формат десятичного файла.");
}


static void LoadMKB(FILE *f, long sz)
{ if (sz==4096)
  { size_t n = fread(bR, 1, sz, f);
    if ((n!=sz) || ferror(f))
    { LoadFailed("двоичных регистров", n);  
      return; 
    }
    puts("Загружено 4096 двоичных регистров.");
  } else OsAlert("vk6", "Неверный формат двоичного файла.");
}


static void LoadMKT(FILE *f, long sz)
{ if (sz==3072)
  { size_t n = fread(bR+4096, 1, sz, f);
    if ((n!=sz) || ferror(f))
    { LoadFailed("двоичных регистров", n);  
      return; 
    }
    puts("Загружен текст.");
  } else OsAlert("vk6", "Неверный формат файла текста.");
}


static void LoadMKN(FILE *f, long sz)
{ OsAlert("vk6", "Загрузка файлов MKN не предусмотрена.");
}


static void LoadMKI(FILE *f, long sz)
{ flgLoadedMKI = 0;
  if (sz==1028)
  { unsigned char s[4];
    size_t n = fread(s, 1, 4, f);
    if ((n==4) && (ferror(f)==0) && (s[2]==128) && (s[3]==64))
    { n = fread(bScreen, 1, 1024, f);
      if ((n!=1024) || ferror(f))
      { LoadFailed("байт", n);  
        return; 
      }
      flgLoadedMKI = 1;
      puts("Загружен графический экран.");
    } else OsAlert("vk6", "Неверный формат файла экрана.");
  } else OsAlert("vk6", "Неверный формат файла экрана.");
}


static void LoadFile(const unsigned char *s)
{ FILE *f;
  size_t sl=strlen((char *)s);
  if (sl && ((f=fopen((char *)s,"rb"))!=NULL))
  { long sz;
    if (fseek(f,0,SEEK_END) || ((sz=ftell(f))==-1) || fseek(f,0,SEEK_SET))
      printf("Не могу определить длину файла '%s'.\n", s);
    else switch(tolower(s[sl-1]))
    { case 'p':
        printf("Загружаю файл программы '%s' длиной %ld байт.\n", s, sz);
        LoadMKP(f, sz);
        break;
      case 'l':
        printf("Загружаю файл исходного текста '%s' длиной %ld байт.\n",
               s, sz);
        LoadMKL(f, sz);
        break;
      case 'd':
        printf("Загружаю файл десятичных данных '%s' длиной %ld байт.\n",
               s, sz);
        LoadMKD(f, sz);
        break;
      case 'b':
        printf("Загружаю файл двоичных данных '%s' длиной %ld байт.\n", s, sz);
        LoadMKB(f, sz);
        break;
      case 't':
        printf("Загружаю файл текста '%s' длиной %ld байт.\n", s, sz);
        LoadMKT(f, sz);
        break;
      case 'n':
        printf("Загружаю файл блокнота '%s' длиной %ld байт.\n", s, sz);
        LoadMKN(f, sz);
        break;
      case 'i':
        printf("Загружаю файл графического экрана '%s' длиной %ld байт.\n",
               s, sz);
        LoadMKI(f, sz);
        break;
      default:
        printf("Не могу загрузить файл неизвестного типа '%s'.\n", s);
        break;
    }
    fclose(f);
  } else printf("Не могу открыть файл '%s'.\n", s);
}


static void DoRun(void)
{ puts("Исполняю...");
  Run(1);
  puts("Исполнение завершено!");
}


void PutCharVK(int c, int c0)
{ if ((c>=0x20) && (c<0x7f)) putchar(c);
  else if ((c>=0x80) && (c<=0x9f)) printf("%s", alRusUp[c-0x80]);
  else if ((c>=0xa0) && (c<=0xaf)) printf("%s", alRusL1[c-0xa0]);
  else if ((c>=0xb0) && (c<=0xcf))
  { if (alSpec[c-0xb0]) printf("%s", alSpec[c-0xb0]);
    else putchar(c0);
  } else if ((c>=0xe0) && (c<=0xef)) printf("%s", alRusL2[c-0xe0]);
  else switch (c)
  { case 0x00: printf("%s", "▌");         break;
    case 0x10: printf("%s", "►");         break;
    case 0x13: printf("%s", "");         break;
    case 0x15: printf("%s", "↑");         break;
    case 0x16: printf("%s", "↓");         break;
    case 0x17: printf("%s", "→");         break;
    case 0x18: printf("%s", "←");         break;
    case 0x1f: printf("%s", " ");         break;
    case 0x7f: printf("%s", "⌂");         break;
    case 0xf0: printf("%s", alRusYo[0]);  break;
    case 0xf1: printf("%s", alRusYo[1]);  break;
    case 0xfb: printf("%s", "√");         break;
    default: if (c0) putchar(c0);
  }
}


void PutStringVK(const unsigned char *s, int n, int field)
/* Выводит строку s в кодировке ВК-6.
   Максимальная длина строки равна n.
   После строки выводятся пробелы, чтобы было не меньше field символов.
 */
{ int i=0;
  while ((i<n) && s[i]) PutCharVK(s[i++], ' ');
  while (i<field) putchar(' '), i++;
}


static int FindLetter(const char *al[], const unsigned char *s,
                      int *pl)
{ const unsigned char *strLetter = (unsigned char *) al[0];
  int i = 0;
  do
  { *pl = -1;
    do if (strLetter[++(*pl)]=='\0') return i;
    while (strLetter[*pl]==s[*pl]);
    strLetter = (unsigned char *) al[++i];
  } while (strLetter);
  return -1;
}


static void StringToVK(const unsigned char *s, unsigned char *d, int n)
{ int i=0, l, c;
  while ((i<n-1) && *s)
  { c = FindLetter(alRusUp, s, &l);
    if (c>=0) { d[i++]=c+0x80; s+=l; continue; }
    c = FindLetter(alRusL1, s, &l);
    if (c>=0) { d[i++]=c+0xa0; s+=l; continue; }
    c = FindLetter(alRusL2, s, &l);
    if (c>=0) { d[i++]=c+0xe0; s+=l; continue; }
    c = FindLetter(alRusYo, s, &l);
    if (c>=0) { d[i++]=c+0xf0; s+=l; continue; }
    d[i++]=*s++;
  }
  d[i]='\0';
}


void PrintStack(void)
/* Вывести экран режима F АВТ. */
{ if (optClear) putchar('\f');
  switch (flgGrad)
  { case  1: printf("%s", flgLat? " R  " : " Р  "); break;
    case  2: printf("%s", flgLat? " GRD" : " ГРД"); break;
    case  0: printf("%s", flgLat? " G  " : " Г  "); break;
  }
  printf("           ");
  if (iMode==1) PutCharVK('\x10',0); else PutCharVK('\x13',0);
  printf("   %04d\n", iPC);
  puts("-----------------------");
  printf("T  %-20.14Lg\n", RT);
  printf("Z  %-20.14Lg\n", RZ);
  printf("Y  %-20.14Lg\n", RY);
  if (flgError) puts("X          ERROR");
  else printf("X  %-20.14Lg\n", RX);
  if (optX1) printf("X1 %-20.14Lg\n", RX1);
  putchar(' ');
  PutStringVK(strComment, iCommentLen, 23);
  putchar('\n');
}


static int Before(int p1, int p2)
/* Проверка указателей программного кода. */
{ int d = p2-p1;
  if (d<=0) d+=10000;
  return d<50;
}


static void ScreenPROG(void)
{ int pc = iPC-5, i;
  int p = pc-8;
  if (pc<0) pc += 10000;
  if (p<0) p += 10000;
  if (optClear) putchar('\f');
  while (Before(p, pc))
    Decomp(&p, 1);
  for (i = 0; i<6; i++)
  { printf("%04d %02x ", pc, bProg[pc]);
    if (p==pc) PutStringVK(Decomp(&p, 1), 11, 11);
    else printf("           ");
    putchar(';');
    PutCharVK(bProg[pc], 0);
    putchar('\n');
    pc++;
    if (pc>9999) pc -= 10000;
  }
  puts("-----------------------");
}


void PrintFPRG(void)
/* Вывести экран режима F ПРГ. */
{ ScreenPROG();
  printf("        ");
  PutStringVK(strComment, iCommentLen, 14);
  putchar('\n');
}


void PrintPPRG(void)
/* Вывести экран режима Р ПРГ. */
{ ScreenPROG();
  printf("HEX  ");
  PutStringVK(strComment, iCommentLen, 18);
  putchar('\n');
}


void PrintScreen(void)
/* Отобразить неграфический экран. */
{ switch (iMode)
  { case 0:
    case 1: PrintStack();  break;
    case 2: PrintFPRG();   break;
    case 3: PrintPPRG();   break;
  }
}


static void Dialog(void)
/* Простейший интерфейс "командная строка". */
{ static char s[81];
  static unsigned char d[81];
  flgExit = 0;
  flgDisplay = 1;
  do
  { if (flgDisplay) PrintScreen();
    flgDisplay = 1;
    putchar(':');
    fgets(s, 80, stdin);
    StringToVK((unsigned char *)s, d, 80);
    Execute(d);
  } while (!flgExit);
}


void DrawScreen(void)
/* Отобразить графический экран. */
{ int s,b,x;
  if (optClear) putchar('\f');
  if (optRot) /* 128x32 */
  { for (s=0; s<1024; s+=128)
      for (b=0; b<8; b+=2)
      {  for (x=0; x<128; x++)
         { int b1 = (bScreen[s+x]>>b)&3;
           printf("%s", alBlocks[optNegativ][b1]);
         }
         putchar('\n');
      }
  } else /* 64x64 */
  { for (x=126; x>=0; x-=2)
    { for (s=0; s<1024; s+=128)
        for (b=0; b<8; b++)
        { int b1 = (bScreen[s+x]>>b)&1, b2 = (bScreen[s+x+1]>>b)&1;
          printf("%s", alBlocks[optNegativ][b1*2+b2]);
        }
      putchar('\n');
    }
  }
}


int main(int argc, const unsigned char *argv[])
{ int i;
  iExitCode = 0;
  puts("Калькулятор ВК-6 v" STR_VERSION);
  TurnOn(PrintStack, DrawScreen);
  for (i = 1; i<argc; i++)
    if (argv[i][0]=='-') Options(argv[i]); else LoadFile(argv[i]);
  if (optDialog) Dialog(); else if (flgLoadedMKP) DoRun(), PrintScreen();
  TurnOff();
  return iExitCode;
}

